from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing_extensions import Self
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypeVar,
    Union as _Union,
)
import itertools
import datetime as _dt
from ...util.source import SourcePos

from pandas import DataFrame

if TYPE_CHECKING:
    from .metamodel_analyzer import Analysis

#------------------------------------------------------
# Immutable Sequence (ISeq)
#------------------------------------------------------

T = TypeVar("T")
ISeq = tuple[T, ...]

#-----------------------------------------------------------------------------
# ID / Base Node
#-----------------------------------------------------------------------------

_id_counter = itertools.count(1)

def _next_id() -> int:
    return next(_id_counter)

@dataclass(repr=False, frozen=True, slots=True)
class Node():
    id: int = field(default_factory=_next_id, compare=False, hash=False, kw_only=True)
    source: SourcePos|None = field(default=None, compare=False, hash=False, kw_only=True)
    annotations: tuple[Annotation,...] = field(default=(), compare=False, hash=False, kw_only=True)

    def __str__(self) -> str:
        try:
            from .pprint import format
            return format(self)
        except Exception as e:
            print(e)
            return object.__repr__(self)

    __repr__ = __str__

    def mut(self: Self, **changes) -> Self:
        """Convenience to return a modified copy, but keeping the id."""
        changes["id"] = self.id
        return replace(self, **changes)

    def replace(self: Self, **changes) -> Self:
        """Convenience to return a modified copy, with a new id."""
        changes["id"] = _next_id()
        return replace(self, **changes)

#-----------------------------------------------------------------------------
# Capabilities, Reasoners
#-----------------------------------------------------------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Capability(Node):
    name: str = ""

@dataclass(repr=False, frozen=True, slots=True)
class Reasoner(Node):
    type: str = ""  # e.g., "SQL", "Logic", "ML", "MathOpt"
    info: Any = None
    capabilities: ISeq[Capability] = field(default=())
    relations: ISeq[Relation] = field(default=())

# -----------------------------------------------------------------------------
# Relation
# -----------------------------------------------------------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Field(Node):
    name: str = ""
    type: Type = field(default_factory=lambda:None) #type: ignore
    input: bool = False  # True if value must be supplied (e.g., parameter/input)
    _relation : Relation = field(default=None, compare=False, hash=False, kw_only=True) #type: ignore

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Field):
            return False
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

@dataclass(repr=False, frozen=True, slots=True)
class Reading(Node):
    name: str = field(default="")
    # for a reading like "{Foo} has {bar:int} and {baz}",
    # parts would be [0, " has ", 1, " and ", 2]
    parts: ISeq[str|int] = field(default=())

    @property
    def field_order(self) -> list[int]:
        return [p for p in self.parts if isinstance(p, int)]

@dataclass(repr=False, frozen=True, slots=True)
class Overload(Node):
    types: ISeq[Type] = field(default=())

@dataclass(repr=False, frozen=True, slots=True)
class Relation(Node):
    name: str = ""
    fields: ISeq[Field] = field(default=())
    requires: ISeq[Capability] = field(default=())
    readings: ISeq[Reading] = field(default=())
    overloads: ISeq[Overload] = field(default=())

    def __post_init__(self):
        # back-reference fields to this relation
        for f in self.fields:
            object.__setattr__(f, '_relation', self)

    def _dangerous_set_readings(self, readings: ISeq[Reading]) -> None:
        object.__setattr__(self, 'readings', readings)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Relation):
            return False
        return self.id == other.id

@dataclass(repr=False, frozen=True, slots=True)
class UnresolvedRelation(Relation):
    """A placeholder for a relation that could not be resolved during compilation."""
    pass

#------------------------------------------------------
# Annotation
#------------------------------------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Annotation(Node):
    relation: Relation = field(default_factory=Relation)
    args: ISeq[Value] = field(default=())

#-----------------------------------------------------------------------------
# Types
#-----------------------------------------------------------------------------

class TypeNode(Relation):
    """Marker base class for all type nodes."""

    def __post_init__(self):
        if not self.fields:
            object.__setattr__(self, 'fields', (Field(name=self.name, type=self),))
        super().__post_init__()

@dataclass(repr=False, frozen=True, slots=True)
class ScalarType(TypeNode):
    super_types: ISeq[ScalarType] = field(default=())
    identify_by: ISeq[Relation] = field(default=())

    def _force_identify_by(self, identify_by: ISeq[Relation]) -> None:
        object.__setattr__(self, 'identify_by', tuple(identify_by))

    def __hash__(self) -> int:
        return hash(self.id)

@dataclass(repr=False, frozen=True, slots=True)
class NumberType(ScalarType):
    precision: int = 38
    scale: int = 14

    def __hash__(self) -> int:
        return hash((self.precision, self.scale))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, NumberType):
            return False
        return self.precision == other.precision and self.scale == other.scale

@dataclass(repr=False, frozen=True, slots=True)
class UnionType(TypeNode):
    types: ISeq[TypeNode] = field(default=())

    def __post_init__(self):
        object.__setattr__(self, "types", tuple(sorted(self.types, key=lambda t: hash(t))))

    def __hash__(self) -> int:
        return hash(self.types)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, UnionType):
            return False
        return len(self.types) == len(other.types) and all(any(t1 == t2 for t2 in other.types) for t1 in self.types)

@dataclass(repr=False, frozen=True, slots=True)
class ListType(TypeNode):
    element_type: TypeNode = field(default_factory=TypeNode)

    def __hash__(self) -> int:
        return hash((self.element_type))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ListType):
            return False
        return self.element_type == other.element_type

@dataclass(repr=False, frozen=True, slots=True)
class TupleType(TypeNode):
    element_types: ISeq[TypeNode] = field(default=())

    def __hash__(self) -> int:
        return hash(self.element_types)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TupleType):
            return False
        return self.element_types == other.element_types

# Type alias to mirror the spec's union
Type = TypeNode
RelationType = ScalarType(name="Relation")
FieldType = ScalarType(name="Field")
NoneType = ScalarType(name="None")

#------------------------------------------------------
# Table
#------------------------------------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Table(ScalarType):
    columns: ISeq[Relation] = field(default=())
    uri: str = ""

    def __hash__(self) -> int:
        return hash(self.id)

    def _force_columns(self, columns: ISeq[Relation]) -> None:
        object.__setattr__(self, 'columns', tuple(columns))

#------------------------------------------------------
# Data
#------------------------------------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Data(Table):
    """A table of data"""
    data: DataFrame = field(default_factory=lambda: DataFrame([]))
    uri: str = "dataframe://in-memory"

    def __hash__(self) -> int:
        return hash(id(self.data))

# -----------------------------------------------------------------------------
# Values
# -----------------------------------------------------------------------------

Primitive = _Union[str, int, float, bool, _dt.datetime, None]

@dataclass(repr=False, frozen=True, slots=True)
class Var(Node):
    type: Type = field(default_factory=TypeNode)
    name: str = ""

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Var):
            return False
        return self.id == other.id

@dataclass(repr=False, frozen=True, slots=True)
class Literal(Node):
    type: Type = field(default_factory=TypeNode)
    value: Primitive = None


# Value is recursive; allow nested lists of Value
Value = _Union[Var, Literal, Type, Relation, Field, ISeq["Value"], None]

# -----------------------------------------------------------------------------
# Tasks (process algebra)
# -----------------------------------------------------------------------------

@dataclass(repr=False, frozen=True, kw_only=True)
class Task(Node):
    reasoner: Optional[Reasoner] = None

@dataclass(repr=False, frozen=True, slots=True, kw_only=True)
class Container(Task):
    optional:bool = field(default=False)
    scope:bool = field(default=False)

# -----------------------------
# Control Flow
# -----------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Logical(Container):
    body: ISeq[Task] = field(default=())

@dataclass(repr=False, frozen=True, slots=True)
class Sequence(Container):
    tasks: ISeq[Task] = field(default=())

@dataclass(repr=False, frozen=True, slots=True)
class Union(Container):
    tasks: ISeq[Task] = field(default=())

@dataclass(repr=False, frozen=True, slots=True)
class Match(Container):
    tasks: ISeq[Task] = field(default=())

@dataclass(repr=False, frozen=True, slots=True)
class Until(Container):
    check: Task = field(default_factory=Logical)
    body: Task = field(default_factory=Logical)

@dataclass(repr=False, frozen=True, slots=True)
class Wait(Container):
    check: Task = field(default_factory=Logical)

@dataclass(repr=False, frozen=True, slots=True)
class Loop(Container):
    over: ISeq[Var] = field(default=())
    body: Task = field(default_factory=Logical)
    concurrency: int = 1

@dataclass(repr=False, frozen=True, slots=True)
class Break(Task):
    check: Task = field(default_factory=Logical)

# -----------------------------
# Constraints
# -----------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Require(Container):
    domain: Task = field(default_factory=Logical)
    check: Task = field(default_factory=Logical)
    error: Optional[Task] = None

# -----------------------------
# Quantifiers
# -----------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Not(Container):
    task: Task = field(default_factory=Logical)

@dataclass(repr=False, frozen=True, slots=True)
class Exists(Container):
    vars: ISeq[Var] = field(default=())
    task: Task = field(default_factory=Logical)

# -----------------------------
# Relation Ops
# -----------------------------

class Effect(str, Enum):
    derive = "derive"
    insert = "insert"
    delete = "delete"

@dataclass(repr=False, frozen=True, slots=True)
class Lookup(Task):
    relation: Relation = field(default_factory=Relation)
    args: ISeq[Value] = field(default=())
    reading_hint: Optional[Reading] = None

@dataclass(repr=False, frozen=True, slots=True)
class Update(Task):
    relation: Relation = field(default_factory=Relation)
    args: ISeq[Value] = field(default=())
    effect: Effect = Effect.derive
    reading_hint: Optional[Reading] = None

@dataclass(repr=False, frozen=True, slots=True)
class Aggregate(Task):
    aggregation: Relation = field(default_factory=Relation)
    projection: ISeq[Var] = field(default=())
    group: ISeq[Var] = field(default=())
    args: ISeq[Value] = field(default=())
    body: Task = field(default_factory=Logical)

@dataclass(repr=False, frozen=True, slots=True)
class Construct(Task):
    values: ISeq[Value] = field(default=())
    id_var: Var = field(default_factory=Var)


# -----------------------------------------------------------------------------
# Model
# -----------------------------------------------------------------------------

@dataclass(repr=False, frozen=True, slots=True)
class Model(Node):
    reasoners: ISeq[Reasoner] = field(default=())
    relations: ISeq[Relation] = field(default=())
    types: ISeq[Type] = field(default=())
    root: Task = field(default_factory=Logical)
    _analysis: Analysis = None # type: ignore - will be set by Analyzer

# -----------------------------------------------------------------------------
# Public exports
# -----------------------------------------------------------------------------

__all__ = [
    # Base
    "Node",
    # Capabilities / Reasoners
    "Capability",
    "Reasoner",
    # Types
    "TypeNode",
    "ScalarType",
    "NumberType",
    "UnionType",
    "ListType",
    "Type",
    # Relations
    "Field",
    "Reading",
    "Relation",
    "Annotation",
    # Values
    "Primitive",
    "Var",
    "Literal",
    "Value",
    "Var",
    # Tasks
    "Task",
    # Control flow
    "Container",
    "Logical",
    "Sequence",
    "Union",
    "Match",
    "Until",
    "Wait",
    "Loop",
    "Break",
    # Constraints
    "Require",
    # Quantifiers
    "Not",
    "Exists",
    # Relation ops
    "Effect",
    "Lookup",
    "Update",
    "Aggregate",
    "Construct",
    "Data",
    # Model
    "Model",
]



from v0.relationalai.semantics import Model, Integer, define, select
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.path_algorithms.one_sided_ball_repetition import ball_with_repetition


# Test with grid graph
model_grid = Model("tests_one_sided_ball_repetition grid", dry_run=False)

grid = Graph.construct_grid(model_grid, 3)

n = Integer.ref()
u = grid.Node.ref()

source_1 = grid.Node.new(row = 1, col = 2)

Source_1 = model_grid.Concept("Source_1", extends=[grid.Node])

define(Source_1(source_1))

source_ball_1 = ball_with_repetition(grid, Source_1, 3)
iter_set_1 = select(n, u.row, u.col).where(source_ball_1(n, u)).to_df()
set_ball_1 = set(row for row in iter_set_1.itertuples(index = False, name = None))

expected_ball_1 = {
    (0, 1, 2), (1, 2, 2), (1, 1, 3), (2, 3, 2), (2, 2, 3), (3, 3, 3)
}

assert set_ball_1 == expected_ball_1


# First test with diamond graph
model_diamond = Model("tests_one_sided_ball_repetition diamond", dry_run=False)

diamond = Graph.construct_diamond(model_diamond, 3)

v = diamond.Node.ref()

source_2 = diamond.Node.new(id = 2)

Source_2 = model_diamond.Concept("Source_2", extends=[diamond.Node])

define(Source_2(source_2))

source_ball_2 = ball_with_repetition(diamond, Source_2, 4)
iter_set_2 = select(n, v.id).where(source_ball_2(n, v)).to_df()
set_ball_2 = set(row for row in iter_set_2.itertuples(index = False, name = None))

expected_ball_2 = {
    (0, 2), (1, 4), (2, 5), (2, 6), (3, 7), (4, 8), (4, 9)
}

assert set_ball_2 == expected_ball_2


# Second test with diamond graph
source_ball_3 = ball_with_repetition(diamond, Source_2, 0)
iter_set_3 = select(n, v.id).where(source_ball_3(n, v)).to_df()
set_ball_3 = set(row for row in iter_set_3.itertuples(index = False, name = None))

expected_ball_3 = {
    (0, 2)
}

assert set_ball_3 == expected_ball_3


from datetime import date, datetime
from decimal import Decimal as PyDecimal
from types import NoneType
import numpy as np
import pandas as pd

#--------------------------------------------------
# Python types
#--------------------------------------------------

pytype_to_concept_name: dict[object, str] = {
    NoneType:  "Any",
    int: "Integer", float: "Number(38,14)", str: "String", bool: "Boolean",
    date: "Date", datetime: "DateTime", PyDecimal: "Number(38,14)",
    # Int128Dtype(): "Integer",

    # NumPy dtypes
    **{np.dtype(k): "Integer" for k in ("int8","int16","int32","int64","uint8","uint16","uint32","uint64")},
    **{np.dtype(k): "Number(38,14)" for k in ("float32","float64")},
    np.dtype("bool"): "Boolean",
    np.dtype("object"): "String",
    np.dtype("datetime64[ns]"): "DateTime",
    np.dtype("datetime64[ms]"): "DateTime",
    np.dtype("datetime64[s]"): "DateTime",

    # Pandas extension dtypes
    **{t(): "Integer" for t in (pd.Int8Dtype, pd.Int16Dtype, pd.Int32Dtype, pd.Int64Dtype,
                        pd.UInt8Dtype, pd.UInt16Dtype, pd.UInt32Dtype, pd.UInt64Dtype)},
    **{t(): "Number(38,14)" for t in (pd.Float32Dtype, pd.Float64Dtype)},
    pd.StringDtype(): "String",
    pd.BooleanDtype(): "Boolean",
    "int": "Integer", "float": "Number(38,14)", "str": "String", "bool": "Boolean",
    "date": "Date", "datetime": "DateTime", "decimal": "Number(38,14)",
    "datetime.date": "Date", "datetime.datetime": "DateTime"
}



import warnings

from v0.relationalai.semantics.metamodel.dependency import analyze_bindings

__all__ = ["analyze_bindings"]

warnings.warn(
    "relationalai.early_access.metamodel.dependency is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.dependency",
    DeprecationWarning,
    stacklevel=2,
)


from relationalai.semantics.frontend.base import Field, Library

library = Library("lqp")

Symbol = library.Type("Symbol")

export = library.Relation("export", [])
external = library.Relation("external", [])
from_cdc = library.Relation("from_cdc", [])
adhoc = library.Relation("adhoc", [])
track = library.Relation("track", [Field.input("library", Symbol), Field.input("relation", Symbol)])


from __future__ import annotations

import base64
import json
from urllib.parse import quote, urlencode
import pyarrow as pa
import requests
from email import message_from_bytes, policy
from email.message import EmailMessage

from dataclasses import dataclass
from typing import Any, Dict, List, Iterable, Literal, Optional, Tuple, Union
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .client import ResourcesBase, ProviderBase
from .config import Config
from .types import TransactionAsyncResponse
from .util import get_pyrel_version
from ..errors import ResponseStatusException
from .. import debugging

@dataclass
class Endpoint:
    method: str
    endpoint: str

TXN_ABORT_REASON_TIMEOUT = "QUERY_TIMEOUT_EXCEEDED"

class LocalClient:

    def __init__(self, host: str, port: int):
        self.service_endpoint = f"http://{host}:{port}"
        self.endpoints: Dict[str, Endpoint] = {
            "create_txn": Endpoint(method="POST", endpoint="/transactions"), # API for creating a (query) transaction
            "create_db": Endpoint(method="POST", endpoint="/v2/<local_org>/<local_account>/databases"), # API for creating a database
        }
        self.http_session = self._create_retry_session()

    def _create_retry_session(self) -> requests.Session:
        http_session = requests.Session()
        retries = Retry(
            total=3,
            backoff_factor=0.3,
            status_forcelist=[500, 502, 503, 504],
            allowed_methods=frozenset({"GET", "POST", "PUT", "DELETE"}),
            raise_on_status=False
        )
        adapter = HTTPAdapter(max_retries=retries)
        http_session.mount("http://", adapter)
        http_session.mount("https://", adapter)
        http_session.headers.update({
            "Connection": "keep-alive",
            "Accept": "application/json",
            "user-agent": get_pyrel_version(None)
        })
        return http_session

    def request(
        self,
        endpoint: str,
        payload: Dict[str, Any] | None = None,
        headers: Dict[str, str] | None = None,
        path_params: Dict[str, str] | None = None,
        query_params: Dict[str, str] | None = None,
    ) -> requests.Response:
        with debugging.span("local_request"):
            url, method = self._prepare_url(endpoint, path_params, query_params)
            request_headers = self._prepare_headers(headers)
            return self.http_session.request(method, url, json=payload, headers=request_headers)

    def _prepare_url(self, endpoint: str, path_params: Dict[str, str] | None = None, query_params: Dict[str, str] | None = None) -> Tuple[str, str]:
        try:
            ep = self.endpoints[endpoint]
        except KeyError:
            raise ValueError(f"Invalid endpoint: {endpoint}. Available endpoints: {list(self.endpoints.keys())}")
        url = f"{self.service_endpoint}{ep.endpoint}"
        if path_params:
            escaped_path_params = {k: quote(v, safe='') for k, v in path_params.items()}
            url = url.format(**escaped_path_params)
        if query_params:
            url += '?' + urlencode(query_params)
        return url, ep.method

    def _prepare_headers(self, headers: Dict[str, str] | None) -> Dict[str, str]:
        request_headers = {}
        if headers:
            request_headers.update(headers)
        return request_headers

class LocalResources(ResourcesBase):

    def __init__(
        self,
        profile: str | None = None,
        config: Config | None = None,
        connection: Any | None = None,
        dry_run: bool = False,
        reset_session: bool = False,
        generation: Any | None = None,
        language: str = "lqp",
    ):
        super().__init__(config=config)
        self.language = language

        host = self.config.get("host", None) or "localhost"
        port = int(self.config.get("port", None) or 8010)
        self.client = LocalClient(host, port)

        # Keep track of whether we already created the local database to run on.
        self._db_created = False

    def reset(self):
        raise NotImplementedError("reset not supported in local mode")
    
    #--------------------------------------------------
    # Check direct access is enabled (0 implemented)
    #--------------------------------------------------

    def is_direct_access_enabled(self) -> bool:
        raise NotImplementedError("is_direct_access_enabled not supported in local mode")

    #--------------------------------------------------
    # Snowflake Account Flags (0 implemented)
    #--------------------------------------------------

    def is_account_flag_set(self, flag: str) -> bool:
        raise NotImplementedError("is_account_flag_set not supported in local mode")

    #--------------------------------------------------
    # Databases (0 implemented)
    #--------------------------------------------------

    def get_database(self, database: str):
        raise NotImplementedError("get_database not supported in local mode")

    def get_installed_packages(self, database: str):
        raise NotImplementedError("get_installed_packages not supported in local mode")

    def _create_database(self, name:str):
        with debugging.span("create_database"):
            response = self.client.request(
                endpoint="create_db",
                payload = {
                    "language": "lqp",
                    "readonly": False,
                    "database_id": name,
                    "transaction_id": "mock_txn_id:create-db",
                }
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to create database in local mode.", response)

            parsed_response = self._parse_txn_response(response)

            state = parsed_response["state"]
            if state != "COMPLETED":
                # NOTE for now we just ignore it when the database creation transaction
                # aborts. In most cases where this happens, it is because the database
                # already exists. If it doesn't, we'll fail later on anyways.
                pass

            return parsed_response

    #--------------------------------------------------
    # Engines (0 implemented)
    #--------------------------------------------------

    def get_version(self):
        raise NotImplementedError("get_version not supported in local mode")

    def get_engine_sizes(self, cloud_provider: str | None = None):
        raise NotImplementedError("get_engine_sizes not supported in local mode")

    def list_engines(self, state: str | None = None):
        raise NotImplementedError("list_engines not supported in local mode")

    def get_engine(self, name: str):
        raise NotImplementedError("get_engine not supported in local mode")

    def get_cloud_provider(self) -> str:
        raise NotImplementedError("get_cloud_provider not supported in local mode")

    def is_valid_engine_state(self, name: str) -> bool:
        raise NotImplementedError("is_valid_engine_state not supported in local mode")

    def create_engine(self, name: str, size:str|None=None, auto_suspend_mins:int|None=None):
        raise NotImplementedError("create_engine not supported in local mode")

    def delete_engine(self, name: str, force: bool = False):
        raise NotImplementedError("delete_engine not supported in local mode")

    def suspend_engine(self, name: str):
        raise NotImplementedError("suspend_engine not supported in local mode")

    def resume_engine(self, name: str, headers: Dict | None = None):
        raise NotImplementedError("resume_engine not supported in local mode")

    def resume_engine_async(self, name: str):
        raise NotImplementedError("resume_engine_async not supported in local mode")

    def alter_engine_pool(self, size: str | None = None, mins: int | None = None, maxs: int | None = None):
        raise NotImplementedError("alter_engine_pool not supported in local mode")

    def auto_create_engine_async(self, name: str | None = None) -> str:
        raise NotImplementedError("auto_create_engine_async not supported in local mode")

    #--------------------------------------------------
    # Graphs (0 implemented)
    #--------------------------------------------------

    def list_graphs(self):
        raise NotImplementedError("list_graphs not supported in local mode")

    def get_graph(self, name: str):
        raise NotImplementedError("get_graph not supported in local mode")

    def create_graph(self, name: str):
        raise NotImplementedError("create_graph not supported in local mode")

    def delete_graph(self, name: str, force=False, language: str = "lqp"):
        # NOTE There is no way to delete a local graph at the moment, which is documented in
        # the README. This function will get called in regular usage, so for now we just
        # silently do nothing.
        return None

    def clone_graph(self, target_name: str, source_name: str, nowait_durable: bool = True):
        raise NotImplementedError("clone_graph not supported in local mode")

    def maybe_poll_use_index(
        self,
        app_name: str,
        sources: Iterable[str],
        model: str,
        engine_name: str,
        engine_size: str | None = None,
        program_span_id: str | None = None,
        headers: Dict | None = None,
    ):
        # Snowflake sources are not supported in local mode.
        if len(list(sources)) != 0:
            raise NotImplementedError("Snowflake sources are not supported in local mode")

        # Nothing to do for local mode if the database was already created.
        if self._db_created:
            return

        self._create_database(model)
        self._db_created = True

    #--------------------------------------------------
    # Models (0 implemented)
    #--------------------------------------------------

    def list_models(self, database: str, engine: str):
        raise NotImplementedError("list_models not supported in local mode")

    def create_models(self, database: str, engine: str, models):
        raise NotImplementedError("create_models not supported in local mode")

    def delete_model(self, database: str, engine: str, name: str):
        raise NotImplementedError("delete_model not supported in local mode")

    def create_models_code(self, models):
        raise NotImplementedError("create_models_code not supported in local mode")

    #--------------------------------------------------
    # Exports (0 implemented)
    #--------------------------------------------------

    def list_exports(self, database: str, engine: str):
        raise NotImplementedError("list_exports not supported in local mode")

    def create_export(self, params):
        raise NotImplementedError("create_export not supported in local mode")

    def create_export_table(self, database: str, engine: str, table: str, relation: str, columns: Dict[str, str], code: str, refresh: str | None = None):
        raise NotImplementedError("create_export_table not supported in local mode")

    def delete_export(self, database: str, engine: str, name: str):
        raise NotImplementedError("delete_export not supported in local mode")

    #--------------------------------------------------
    # Imports (0 implemented)
    #--------------------------------------------------

    def get_import_stream(self, name: str | None, model: str | None):
        raise NotImplementedError("get_import_stream not supported in local mode")

    def change_stream_status(self, stream_id: str, model: str, suspend: bool):
        raise NotImplementedError("change_stream_status not supported in local mode")

    def change_imports_status(self, suspend: bool):
        raise NotImplementedError("change_imports_status not supported in local mode")

    def get_imports_status(self):
        raise NotImplementedError("get_imports_status not supported in local mode")

    def set_imports_engine_size(self, size: str):
        raise NotImplementedError("set_imports_engine_size not supported in local mode")

    def list_imports(self, id: str | None = None, name: str | None = None, model: str | None = None, status: str | None = None, creator: str | None = None):
        raise NotImplementedError("list_imports not supported in local mode")

    def poll_imports(self, sources, model: str):
        raise NotImplementedError("poll_imports not supported in local mode")

    def create_import_stream(self, source, model: str, rate: int | None = None, options: dict | None = None):
        raise NotImplementedError("create_import_stream not supported in local mode")

    def create_import_snapshot(self, source, model: str, options: dict | None):
        raise NotImplementedError("create_import_snapshot not supported in local mode")

    def delete_import(self, import_name: str, model: str, force=False):
        raise NotImplementedError("delete_import not supported in local mode")

    #--------------------------------------------------
    # Exec Async
    #--------------------------------------------------
    
    def _parse_multipart_response(self, response: requests.Response) -> Dict[str, Any]:
        response_map = {}
        response_map['results'] = {}

        content_type = response.headers.get("Content-Type")
        if not content_type:
            raise ValueError("Missing Content-Type header")

        # Parse using Python's email module
        # Construct a proper MIME message with headers
        mime_content = f"Content-Type: {content_type}\r\n\r\n".encode('utf-8') + response.content
        msg: EmailMessage = message_from_bytes(mime_content, policy=policy.default)  # type: ignore[assignment]

        if not msg.is_multipart():
            raise ValueError("Response is not multipart")

        for part in msg.iter_parts():
            part_content_type = part.get_content_type() or 'application/octet-stream'
            part_content: bytes = part.get_payload(decode=True)  # type: ignore[assignment]

            # Extract disposition info
            disposition = part.get('Content-Disposition', '')
            part_name = None
            part_filename = None
            if disposition:
                params = {}
                for param_str in disposition.split(';')[1:]:
                    if '=' in param_str:
                        key, val = param_str.strip().split('=', 1)
                        params[key] = val.strip('"')
                part_name = params.get('name')
                part_filename = params.get('filename')

            if 'json' in part_content_type.lower() and not part_filename:
                # JSON part (no filename)
                try:
                    response_map[part_name] = json.loads(part_content.decode('utf-8'))
                except json.JSONDecodeError as e:
                    raise ValueError(f"Failed to decode JSON part: {e}")

            elif part_filename:
                # File/attachment part
                response_map['results'][part_filename] = {
                    'name': part_name,
                    'filename': part_filename,
                    'content': part_content,
                    'content_type': part_content_type
                }

        return response_map

    def _parse_txn_response(self, get_txn_response: requests.Response) -> Dict[str, Any]:
        with debugging.span("parse_txn_response"):
            response_content_type = get_txn_response.headers.get("Content-Type", "")
            if "application/json" in response_content_type:
                parsed_response = get_txn_response.json()
            elif "multipart/form-data" in response_content_type:
                parsed_response = self._parse_multipart_response(get_txn_response)
                state = parsed_response.get('transaction', {}).get('state', None)
                parsed_response['state'] = state
                txn_id = parsed_response.get('transaction', {}).get('id', None)
                parsed_response['id'] = txn_id
            else:
                raise Exception(f"Unsupported Response Content Type:, {response_content_type}")
            return parsed_response

    def _create_transaction(
        self,
        target_endpoint: str,
        payload: Dict[str, Any] | None = None,
        headers: Dict[str, str] | None = None,
        path_params: Dict[str, str] | None = None,
        query_params: Dict[str, str] | None = None
    ) -> Dict[str, Any]:
        with debugging.span("transaction"):
            response = self.client.request(
                target_endpoint,
                path_params=path_params,
                query_params=query_params,
                payload=payload,
                headers=headers,
            )
            assert response, "No results from create transaction request."
            if response.status_code not in [200, 201]:
                raise ResponseStatusException("Failed to create transaction.", response)

            return self._parse_txn_response(response)

    def _convert_txn_response(self, parsed_response: Dict[str, Any]) -> TransactionAsyncResponse:
        results = []
        for (file_name, file_info) in parsed_response.get("results", {}).items():
            if not file_name.endswith(".arrow"):
                continue
            with pa.ipc.open_stream(file_info["content"]) as reader:
                schema = reader.schema
                batches = [batch for batch in reader]
                table = pa.Table.from_batches(batches=batches, schema=schema)
                results.append({"relationId": file_info["name"], "table": table})

        rsp = TransactionAsyncResponse()
        rsp.transaction = {
            "id": parsed_response["id"],
            "state": parsed_response["state"]
        }
        rsp.results = results
        rsp.metadata = parsed_response.get("metadata", [])
        rsp.problems = parsed_response.get("problems", [])

        return rsp

    def _exec_async_v2(
        self,
        database: str,
        engine: Union[str, None],
        raw_code: str,
        inputs: Dict | None = None,
        readonly=True,
        nowait_durable=False,
        headers: Dict[str, str] | None = None,
        bypass_index=False,
        language: str = "lqp",
        query_timeout_mins: int | None = None,
    ):
        payload = {
            "engine_name": engine,
            "dbname": database,
            "query": raw_code,
            "language": "lqp",
            "readonly": readonly,
        }

        parsed_response = self._create_transaction(
            target_endpoint="create_txn",
            payload=payload,
            headers=headers
        )

        state = parsed_response["state"]
        if state not in ["COMPLETED", "ABORTED"]:
            # Local mode only supports synchronous transactions for now.
            txn_id = parsed_response["id"]
            raise Exception(f"Transaction {txn_id} did not complete or abort synchronously (state = {state}).")

        return self._convert_txn_response(parsed_response)

    #--------------------------------------------------
    # Exec (1 implemented, others are stubs)
    #--------------------------------------------------

    def exec_lqp(
        self,
        database: str,
        engine: str | None,
        raw_code: bytes,
        readonly=True,
        *,
        inputs: Dict | None = None,
        nowait_durable=False,
        headers: Dict | None = None,
        bypass_index=False,
        query_timeout_mins: int | None = None,
    ):
        raw_code_b64 = base64.b64encode(raw_code).decode("utf-8")
        return self._exec_async_v2(
            database, engine, raw_code_b64, inputs, readonly, nowait_durable,
            headers=headers, bypass_index=bypass_index, language='lqp',
            query_timeout_mins=query_timeout_mins,
        )

    def exec_raw(
        self,
        database: str,
        engine: str | None,
        raw_code: str,
        readonly: bool = True,
        *,
        inputs: Dict | None = None,
        nowait_durable: bool = False,
        headers: Dict | None = None,
        query_timeout_mins: Optional[int] = None,
    ):
        raise NotImplementedError("exec_raw not supported in local mode - use exec_lqp instead")

    def exec_format(self, database: str, engine: str, raw_code: str, cols: List[str], format:str, inputs: Dict | None = None, readonly: bool = True, nowait_durable=False, skip_invalid_data=False, headers: Dict | None = None, query_timeout_mins: int | None = None) -> Any:
        raise NotImplementedError("exec_format not supported in local mode")

    def format_results(self, results, task=None):
        raise NotImplementedError("format_results not supported in local mode")

    def to_model_type(self, model, name: str, source: str):
        raise NotImplementedError("to_model_type not supported in local mode")

    def _exec(self, code: str, params=None, raw=False, help=True):
        raise NotImplementedError("_exec not supported in local mode")

    #--------------------------------------------------
    # Transactions (0 implemented)
    #--------------------------------------------------

    def get_transaction(self, transaction_id: str):
        raise NotImplementedError("get_transaction not supported in local mode")

    def list_transactions(self, *, limit: int, only_active=False, **kwargs):
        raise NotImplementedError("list_transactions not supported in local mode")

    def cancel_transaction(self, transaction_id: str):
        raise NotImplementedError("cancel_transaction not supported in local mode")

    def cancel_pending_transactions(self):
        raise NotImplementedError("cancel_pending_transactions not supported in local mode")

    def get_transaction_events(self, transaction_id: str, continuation_token: str):
        raise NotImplementedError("get_transaction_events not supported in local mode")

class LocalProvider(ProviderBase):

    def __init__(
        self,
        host: str = "localhost",
        port: int = 8010,
        profile: str | None = None,
        config: Config | None = None,
        resources: LocalResources | None = None,
    ):
        if resources:
            self.resources = resources
        else:
            self.resources = LocalResources(
                profile=profile,
                config=config,
                connection=None,
                dry_run=False,
                reset_session=False,
                generation=None,
                language="lqp",
            )
        self.config = self.resources.config

    # A mock sql function that simulates SQL execution for the LocalProvider. This is only
    # here to make the test suite pass.
    def sql(self, query: str, params: List[Any] = [], format: Literal["list", "pandas", "polars", "lazy"] = "list"):
        if format == "list":
            # Used by "test_snapshot_abstract.py"
            import re
            pattern = r"\s*SELECT EXISTS\(\s*SELECT 1\s*FROM (\w+)\.INFORMATION_SCHEMA\.TABLES\s*WHERE TABLE_SCHEMA = '(\w+)'\s*AND TABLE_NAME = '(\w+)'\s*\) AS TABLE_EXISTS;\s*"
            query_cleaned = query.replace("\n", " ").replace("  ", " ")
            match_obj = re.match(pattern, query_cleaned, re.DOTALL)
            if match_obj:
                return [{"TABLE_EXISTS": False}]
        else:
            raise NotImplementedError("Only specific SQL queries are defined in local mode.")



from __future__ import annotations

from typing import Literal, Any, TYPE_CHECKING
from pydantic import Field

if TYPE_CHECKING:
    import duckdb

from .base import BaseConnection


class DuckDBConnection(BaseConnection):
    type: Literal["duckdb"] = "duckdb"

    path: str = Field(..., description="Path to DuckDB database file (or ':memory:' for in-memory)")
    read_only: bool = Field(False, description="Open database in read-only mode")
    config: dict[str, Any] | None = Field(None, description="Additional DuckDB configuration options")

    def get_session(self) -> duckdb.DuckDBPyConnection:
        import duckdb

        if self._cached_session is None:
            self._cached_session = duckdb.connect(
                database=self.path,
                read_only=self.read_only,
                config=self.config or {}
            )

        return self._cached_session



import warnings

from v0.relationalai.semantics.lqp.primitives import lqp_avg_op, lqp_operator, build_primitive

__all__ = ["lqp_avg_op", "lqp_operator", "build_primitive"]

warnings.warn(
    "relationalai.early_access.lqp.primitives is deprecated. "
    "Please migrate to relationalai.semantics.lqp.primitives",
    DeprecationWarning,
    stacklevel=2,
)


"""
Error handlers for Snowflake Resources using Strategy Pattern.

Each error handler encapsulates the detection logic and exception creation
for a specific type of error. Handlers are processed in order until one matches.
"""
from __future__ import annotations
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from ....errors import (
    DuoSecurityFailed,
    EngineNotFoundException,
    EnginePending,
    EngineNameValidationException,
    EngineProvisioningFailed,
    EngineResumeFailed,
    RAIAbortedTransactionError,
    RAIException,
    SnowflakeAppMissingException,
    SnowflakeDatabaseException,
    SnowflakeRaiAppNotStarted,
)

if TYPE_CHECKING:
    from .snowflake import ExecContext
    from .snowflake import Resources

from .util import is_database_issue, is_engine_issue, collect_error_messages

class ErrorHandler(ABC):
    """Base class for error handlers using Strategy Pattern."""

    @abstractmethod
    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        """Check if this handler can process the error."""
        pass

    @abstractmethod
    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        """Handle the error and either raise an exception or return an alternate result."""
        pass


class DuoSecurityErrorHandler(ErrorHandler):
    """Handle Duo security authentication errors."""

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        messages = collect_error_messages(error)
        return any("duo security" in msg for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        raise DuoSecurityFailed(error)


class AppMissingErrorHandler(ErrorHandler):
    """Handle missing RAI app database errors."""

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        rai_app = resources.config.get("rai_app_name", "")
        if not isinstance(rai_app, str):
            return False
        pattern = f"database '{rai_app}' does not exist or not authorized."
        messages = collect_error_messages(error)
        return any(re.search(pattern.lower(), msg) for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        rai_app = resources.config.get("rai_app_name", "")
        assert isinstance(rai_app, str), f"rai_app_name must be a string, not {type(rai_app)}"
        raise SnowflakeAppMissingException(rai_app, resources.config.get("role"))


class DatabaseErrorsHandler(ErrorHandler):
    """Handle database-related errors from Snowflake/RAI."""

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        """Check if error is database-related."""
        # Use collect_error_messages to get all messages from exception chain
        messages = collect_error_messages(error)
        return any(is_database_issue(msg) for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        """Handle database errors and raise appropriate exception."""
        raise SnowflakeDatabaseException(error)


class EngineErrorsHandler(ErrorHandler):
    """Handle all engine-related errors from Snowflake/RAI."""

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        """Check if error is engine-related."""
        # Use collect_error_messages to get all messages from exception chain
        messages = collect_error_messages(error)
        return any(is_engine_issue(msg) for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        """Handle engine errors and raise appropriate exception."""
        # Use collect_error_messages to get all messages from exception chain
        messages = collect_error_messages(error)
        engine = resources.get_default_engine_name()
        assert isinstance(engine, str), f"engine must be a string, not {type(engine)}"

        # Check all collected messages for engine error patterns
        for message in messages:
            if "engine is in pending" in message or "engine is provisioning" in message:
                raise EnginePending(engine)
            elif "engine not found" in message or "no engines found" in message:
                raise EngineNotFoundException(engine, str(error))
            elif "engine was deleted" in message:
                raise EngineNotFoundException(engine, "Engine was deleted")
            elif "engine is suspended" in message:
                raise EngineResumeFailed(engine)
            elif "create/resume" in message:
                raise EngineProvisioningFailed(engine, error)

        # Generic engine error - use the original error message
        raise RAIException(str(error))


class ServiceNotStartedErrorHandler(ErrorHandler):
    """Handle RAI service not started errors."""

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        messages = collect_error_messages(error)
        return any(re.search(r"service has not been started", msg) for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        rai_app = resources.config.get("rai_app_name", "")
        assert isinstance(rai_app, str), f"rai_app_name must be a string, not {type(rai_app)}"
        raise SnowflakeRaiAppNotStarted(rai_app)


class TransactionAbortedErrorHandler(ErrorHandler):
    """Handle transaction aborted errors with problem details."""

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        messages = collect_error_messages(error)
        return any(re.search(r"state:\s*aborted", msg) for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        # Use collect_error_messages to get all messages from exception chain
        messages = collect_error_messages(error)
        # Check all collected messages for transaction ID
        for message in messages:
            txn_id_match = re.search(r"id:\s*([0-9a-f\-]+)", message)
            if txn_id_match:
                problems = resources.get_transaction_problems(txn_id_match.group(1))
                if problems:
                    # Extract problem details (handle both dict and object formats)
                    for problem in problems:
                        if isinstance(problem, dict):
                            type_field = problem.get('TYPE')
                            message_field = problem.get('MESSAGE')
                            report_field = problem.get('REPORT')
                        else:
                            type_field = problem.TYPE
                            message_field = problem.MESSAGE
                            report_field = problem.REPORT
                        raise RAIAbortedTransactionError(type_field, message_field, report_field)
        raise RAIException(str(error))


class UseIndexRetryErrorHandler(ErrorHandler):
    """Handle engine/database errors by polling use_index and retrying the execution.

    Intended for UseIndexResources and subclasses. Register this handler *before*
    the standard Database/Engine error handlers.
    """

    def matches(self, error: Exception, message: str, ctx: 'ExecContext', resources: 'Resources') -> bool:
        if ctx.skip_engine_db_error_retry:
            return False
        messages = collect_error_messages(error)
        return any(msg and (is_database_issue(msg) or is_engine_issue(msg)) for msg in messages)

    def handle(self, error: Exception, ctx: 'ExecContext', resources: 'Resources') -> Any | None:
        poll_use_index = getattr(resources, "_poll_use_index", None)
        if not callable(poll_use_index):
            return None

        engine = resources.get_default_engine_name()
        engine_size = resources.config.get_default_engine_size()
        assert isinstance(engine, str), f"engine must be a string, not {type(engine)}"

        model = getattr(resources, "database", "")
        try:
            poll_use_index(
                app_name=resources.get_app_name(),
                sources=resources.sources,
                model=model,
                engine_name=engine,
                engine_size=engine_size,
            )
        except EngineNameValidationException as e:
            raise EngineNameValidationException(engine) from e

        return ctx.re_execute(resources)




from __future__ import annotations

from v0.relationalai.semantics.internal import internal as b

def unique(*args: b.RelationshipFieldRef|b.Relationship|b.RelationshipRef|b.RelationshipReading, model:b.Model|None=None) -> tuple[b.Expression, ...]:
    if not args:
        raise ValueError("unique expects at least 1 argument")
    if len(args) > 1:
        if not all(isinstance(a, b.RelationshipFieldRef) for a in args):
            raise TypeError("Not all items are relationship fields")
        uc = b.Unique(*args, model=model) # type: ignore
    else:
        assert args
        if isinstance(args[0], b.RelationshipFieldRef):
            uc = b.Unique(*args) # type: ignore
        elif isinstance(args[0], (b.Relationship, b.RelationshipRef, b.RelationshipReading)):
            rel = args[0]._relationship if isinstance(args[0], b.RelationshipRef) else args[0]
            if rel._arity() != 2:
                raise ValueError("Only binary relationships are allowed")
            uc = b.Unique(rel[1], model=model)
        else:
            raise TypeError("Unexpected argument type. Only relationship fields or relationships are allowed")
    return uc.to_expressions()

def exclusive(*args: b.Concept|b.Ref, model:b.Model|None=None) -> tuple[b.Expression, ...]:
    return b.Exclusive(*_extract_concepts(*args), model=model).to_expressions()

def anyof(*args: b.Concept|b.Ref, model:b.Model|None=None) -> tuple[b.Expression, ...]:
    return b.Anyof(*_extract_concepts(*args), model=model).to_expressions()

def oneof(*args: b.Concept | b.Ref, model: b.Model | None = None) -> tuple[b.Expression, ...]:
    return exclusive(*args, model=model) + anyof(*args, model=model)

def _extract_concepts(*args: b.Concept|b.Ref) -> list[b.Concept]:
    concepts = []
    for arg in args:
        c = arg
        if isinstance(c, b.Ref):
            c = c._thing
            if not isinstance(c, b.Concept):
                raise TypeError("Only concepts and concept's ref are allowed as arguments")
        concepts.append(arg)
    return concepts




from ..frontend.base import FieldRef, Fragment, Library, Expression, Field, Value, TupleVariable, Variable, Concept, Ref, MetaRef
from ..frontend import core

# the front-end library object
library = Library("constraints")

#------------------------------------------------------
# Constraint Relations
#------------------------------------------------------

_unique = library.Relation("unique", [Field.input("values", core.Any, is_list=True)])
_unique_fields = library.Relation("unique_fields", [Field.input("fields", core.Field, is_list=True)])
_exclusive = library.Relation("exclusive", [Field.input("concepts", core.Relation, is_list=True)])
_anyof =library.Relation("anyof", [Field.input("concepts", core.Relation, is_list=True)])

#------------------------------------------------------
# API
#------------------------------------------------------

def unique(*args: Variable) -> Expression:
    if all(isinstance(arg, FieldRef) for arg in args):
        fields = [MetaRef(arg._resolved) for arg in args] # type: ignore - we know these are FieldRefs from the all above
        return _unique_fields(TupleVariable(fields))
    return _unique(TupleVariable(args))

def exclusive(*args: Concept|Ref) -> Expression:
    concepts = [arg._to_concept() for arg in args]
    return _exclusive(TupleVariable(concepts))

def anyof(*args: Concept|Ref) -> Expression:
    concepts = [arg._to_concept() for arg in args]
    return _anyof(TupleVariable(concepts))

def oneof(*args: Concept | Ref) -> Fragment:
    model = args[0]._model
    return model.where(anyof(*args), exclusive(*args))



import runpy

from v0.relationalai.early_access.dsl.core.types.standard import standard_value_types
from v0.relationalai.early_access.dsl.bindings.legacy import binding_models
from v0.relationalai.early_access.dsl.ontologies import models, constraints
from v0.relationalai.early_access.dsl.serialize.binding_model import BindingModelObject, BindingObject, BoundRole, BoundAttribute, \
    DataTransformerApplication, DataTransformer, EntityMapRef, Pattern
from v0.relationalai.early_access.dsl.serialize.model import ValueType, ModelObject, DataType, EntityType, Role, Reading, SubtypeArrow, \
    Relationship,UniqueConstraint, MandatoryConstraint, Model, Constraints, Concepts

ENTITY_PATTERN_NAME = 'Entity'


def export_models(ontology_paths) -> list[ModelObject]:
    result = []

    # Add standard Data Types
    data_types_list = []
    for vt in standard_value_types.values():
        data_types_list.append(DataType(id=vt.guid(), name=vt.display()))

    # Scan for Models in given paths
    for path in ontology_paths:
        module_globals = runpy.run_path(path)
        value_types_list = []
        entity_types_list = []
        relationships_list = []
        subtype_arrows_list = []
        has_primary = False
        for var in module_globals:
            obj = module_globals[var]
            if isinstance(obj, models.Model):
                if has_primary and obj.is_primary:
                    raise ValueError('Only one primary model per ontology is allowed.')
                if obj.is_primary:
                    has_primary = True

                uniqueness_list = []
                mandatory_list = []

                for vt in obj._value_types.values():
                    decl = vt._params[0].decl
                    value_types_list.append(
                        ValueType(id=vt.guid(), name=vt.display(), data_type=decl.guid(), data_type_name=decl.display()))

                for et in obj._entity_types.values():
                    entity_types_list.append(
                        EntityType(id=et.guid(), name=et.display(), domain=list(map(lambda d: d.display(), et._domain))))

                for rel in obj._relationships.values():
                    rel_roles = []
                    for role in rel._rolemap.values():
                        rel_roles.append(
                            Role(id=role.guid(), name=role.name(), role_player=role.player().guid(),
                                 role_player_name=role.player().display(),
                                 pre_bound_text=role._prefix,
                                 post_bound_text=role._postfix))

                    rel_readings = []
                    for reading in rel.readings():
                        rel_readings.append(Reading(id=reading.guid(), text=reading.template(),
                                                    roles=list(map(lambda r: r.guid(), reading.roles))))

                    relationships_list.append(
                        Relationship(id=rel.guid(), name=rel._name(), roles=rel_roles, readings=rel_readings,
                                     is_subtype=rel._is_subtype))

                for c in obj._constraints:
                    if isinstance(c, constraints.Unique):
                        uniqueness_list.append(
                            UniqueConstraint(id=c.guid(), roles=list(map(lambda r: r.guid(), c.roles())),
                                             is_preferred_identifier=c.is_preferred_identifier))
                    elif isinstance(c, constraints.Mandatory):
                        mandatory_list.append(MandatoryConstraint(id=c.guid(), roles=[c.role.guid()]))

                for a in obj._subtype_arrows:
                    subtype_arrows_list.append(
                        SubtypeArrow(id=a.guid(), start=a.start.guid(), end=a.end.guid(), start_name=a.start.display(),
                                     end_name=a.end.display()))

                result.append(ModelObject(model=Model(id=obj.guid(), name=obj.name, is_primary=obj.is_primary,
                                                      concepts=Concepts(value_types=value_types_list,
                                                                        entity_types=entity_types_list),
                                                      relationships=relationships_list,
                                                      constraints=Constraints(uniqueness=uniqueness_list,
                                                                              mandatory=mandatory_list),
                                                      data_types=data_types_list,
                                                      subtype_arrows=subtype_arrows_list)))
    return result


def export_model_json(ontology: list[ModelObject]) -> str:
    return ModelObject.schema().dumps(ontology, many=True) # type: ignore


def export_binding_model(binding_model_path) -> BindingModelObject:
    module_globals = runpy.run_path(binding_model_path)
    binding_list = []
    data_transformer_list = []
    # Reification pattern support is a TODO
    pattern_list = []
    entity_map_list = []
    for var in module_globals:
        obj = module_globals[var]
        if isinstance(obj, binding_models.BindingModel):
            for b in obj._bindings:
                ro = b.role
                col = b.column
                attr = BoundAttribute(
                    name=col.name,
                    source=col.part_of.name,
                    hasValue=b.has_value,
                ) if col is not None else None
                bo = BindingObject(
                    id=b.guid(),
                    role=BoundRole(
                        player=ro.player_type._name,
                        relationship=ro.part_of._name(),
                        entityMapRef=b.entity_map.name if b.entity_map is not None else None),
                    attribute=attr,
                )
                # Filter out potential None values then flatten the list first if it becomes (single) nested
                b._applied_transformations = list(filter(lambda x: x is not None, b._applied_transformations))
                b._applied_transformations = [item for sublist in b._applied_transformations for item in sublist]
                if b._applied_transformations and len(b._applied_transformations) > 0:
                    bo.transformations = list(
                        map(
                            lambda t: DataTransformerApplication(transformer=t._relname),
                            b._applied_transformations
                        )
                    )
                binding_list.append(bo)
            data_transformer_list = list(map(
                lambda kv: DataTransformer(
                    name=kv[0],
                    sourceType=kv[1][0]._name.upper(),
                    targetType=kv[1][1]._name.upper()
                ),
                obj._transformers.items()
            ))
            entity_map_list = list(map(
                lambda kv: EntityMapRef(
                    name=kv.name,
                    source=kv.source.name,
                    entityType=kv.entity_type._name,
                    relation=kv.relation.qualified_name()
                ),
                obj._entity_maps.values()
            ))
            pattern_list = list(map(
                lambda kv: Pattern(
                    conceptName=kv[0].__name__,
                    # We use a little different naming convention for entity reificaiton pattern
                    pattern=kv[1].__name__.replace(EntityType.__name__, ENTITY_PATTERN_NAME)
                ),
                obj._patterns.items()
            ))
    return BindingModelObject(bindings=binding_list, patterns=pattern_list, dataTransformers=data_transformer_list,
                              entityMaps=entity_map_list)


def export_binding_model_json(binding_model: BindingModelObject) -> str:
    return BindingModelObject.schema().dumps(binding_model) # type: ignore



import os
import re
from nicegui import ui
import json
import dataclasses


#--------------------------------------------------
# Data
#--------------------------------------------------
last_mod_time = None
current_content: dict[str, str] = dict()

def parse_file(jsonl_content):
    content = dict()
    lines = jsonl_content.strip().split('\n')
    for line in lines:
        data = json.loads(line)
        content[data['id']] = data['content']
    return content

def poll():
    global last_mod_time
    global current_content
    # Check the last modification time of the file
    try:
        mod_time = os.path.getmtime('typer_debug.jsonl')
        if last_mod_time is None or mod_time > last_mod_time:
            last_mod_time = mod_time
            # File has changed, read and parse the new content
            with open('typer_debug.jsonl', 'r') as file:
                content = file.read()
                if content:
                    current_content = parse_file(content)
                    view.refresh()
    except FileNotFoundError:
        pass

#--------------------------------------------------
# UI
#--------------------------------------------------

@ui.refreshable
def view():
    with ui.column():
        # model
        with ui.row():
            with ui.column():
                ui.label("Model").style("font-weight: bold;")
                code(current_content.get('model', ''), language='metamodel')
            with ui.column():
                ui.label("Typed Model").style("font-weight: bold;")
                code(current_content.get('typed_model', ''), language='metamodel')
        # typer
        with ui.row():
            ui.label("Type Propagation Network").style("font-weight: bold;")
            ui.mermaid(current_content.get('model_net', ''),
                       {"maxEdges":10000}
            ).style("display:flex; padding:0px 0px 0px 15px; gap:30px; overflow: auto;").on('error', lambda e: print(e.args['message']))
        # query
        with ui.row():
            with ui.column():
                ui.label("Query").style("font-weight: bold;")
                code(current_content.get('query', ''), language='rel')
            with ui.column():
                ui.label("Typed Query").style("font-weight: bold;")
                code(current_content.get('typed_query', ''), language='rel')

        # typer
        with ui.row():
            ui.label("Type Propagation Network").style("font-weight: bold;")
            ui.mermaid(current_content.get('query_net', ''), {"maxEdges":10000}).style("display:flex; padding:0px 0px 0px 15px; gap:30px; overflow: auto;")

def code(c, language="python"):
    if language == "rel":
        language = "ruby"
    if language == "lqp":
        language = "clojure"
    c = re.sub(r"→", "->", c)
    c = re.sub(r"⇑", "^", c)
    return ui.code(c, language=language).style("border:none; margin:0; padding-right: 30px; ").classes("w-full")

def main(host="0.0.0.0", port=8082):
    global checkboxes
    global toggles
    ui.dark_mode(True)
    view()

    ui.timer(1, poll)
    ui.run(reload=False, host=host, port=port)

if __name__ in {"__main__", "__mp_main__"}:
    main()



"""
Helpers to analyze the metamodel IR.
"""
from __future__ import annotations

import re
from dataclasses import fields
from typing import cast, Tuple, Iterable, Optional, TypeVar
from v0.relationalai.semantics.metamodel import ir, visitor, builtins, types, factory as f
from v0.relationalai.semantics.metamodel.util import NameCache, OrderedSet, FrozenOrderedSet, flatten_tuple, ordered_set



#--------------------------------------------------
# Name helpers
#--------------------------------------------------

def sanitize(name:str) -> str:
    """ Cleanup the name to make it more palatable to names. """
    x = re.sub(r"[ ,\.\(\)\|]", "_", name)
    return x[0:-1] if x[-1] == "_" else x

#--------------------------------------------------
# Checks
#--------------------------------------------------

def is_concept_lookup(node: ir.Lookup|ir.Relation):
    """ Whether this task is a concept lookup. """
    if isinstance(node, ir.Lookup) and is_concept_lookup(node.relation):
        return True
    return builtins.concept_relation_annotation in node.annotations

def is_external(relation: ir.Relation):
   """ Whether this relation is external, by being marked with the external annotation. """
   return builtins.external_annotation in relation.annotations

def is_from_cast(node: ir.Lookup|ir.Relation):
    """ Whether this relation is from cast, by being marked with the from_cast_annotation annotation. """
    if isinstance(node, ir.Lookup) and is_from_cast(node.relation):
        return True
    return builtins.from_cast_annotation in node.annotations

def is_aggregate_input(var: ir.Var, agg: ir.Aggregate):
    """ Whether this var is an input to this aggregation. """
    return (var in agg.args and agg.aggregation.fields[agg.args.index(var)].input)

def is_effective_logical(n: ir.Task):
    """ Whether this task is a Logical and contains an Update child, recursively. """
    return isinstance(n, ir.Logical) and len(visitor.collect_by_type(ir.Update, n)) > 0

def is_nullable_logical(n: ir.Task):
    """ Whether this task is a Logical that contains a hoisted variable with a None default. """
    return isinstance(n, ir.Logical) and any(isinstance(v, ir.Default) and v.value is None for v in n.hoisted)

def relation_is_subtype(r1: ir.Relation, r2: ir.Relation):
    if r1 is r2:
        return True
    if len(r1.fields) != len(r2.fields):
        return False
    return all(types.is_subtype(f1.type, f2.type) for f1, f2 in zip(r1.fields, r2.fields))

def relation_is_proper_subtype(r1: ir.Relation, r2: ir.Relation):
    if r1 is r2:
        return False
    if relation_is_subtype(r1, r2):
        return any(types.is_proper_subtype(f1.type, f2.type) for f1, f2 in zip(r1.fields, r2.fields))
    else:
        return False

def relation_name_prefix(relation: ir.Relation):
    prefix = ""
    if len(relation.fields) > 0 and not is_concept_lookup(relation):
        main_type = relation.fields[0].type
        if (isinstance(main_type, ir.ScalarType) and
            main_type not in types.builtin_types and
            not relation.name.startswith(main_type.name) and
            not relation.name.startswith("_")):
                prefix = f"{main_type.name.lower()}_"
    return prefix

def get_outputs(lookup: ir.Lookup):
    """
        Return an array with the arguments of this lookup that are referring to output fields
        on the relation being looked up.
    """
    if builtins.is_eq(lookup.relation):
        # special case eq because it can be input or output
        x, y = lookup.args[0], lookup.args[1]
        if isinstance(x, ir.Var) and not isinstance(y, ir.Var):
            return [x]
        elif not isinstance(x, ir.Var) and isinstance(y, ir.Var):
            return [y]
        # both are inputs
        return []
    else:
        outputs = []
        # register variables depending on the input flag of the relation bound to the lookup
        for idx, fld in enumerate(lookup.relation.fields):
            arg = lookup.args[idx]
            if isinstance(arg, Iterable):
                # deal with ListType fields that pack arguments in a tuple
                for element in arg:
                    if isinstance(element, ir.Var) and not fld.input:
                        outputs.append(element)
            else:
                if isinstance(arg, ir.Var) and not fld.input:
                    outputs.append(arg)
        return outputs

def get_agg_outputs(lookup: ir.Aggregate):
    """
        Return an array with the arguments of this aggregate that are referring to output fields
        on the relation being looked up.
    """
    outputs = []
    # register variables depending on the input flag of the relation bound to the lookup
    for idx, fld in enumerate(lookup.aggregation.fields):
        arg = lookup.args[idx]
        if isinstance(arg, Iterable):
            # deal with ListType fields that pack arguments in a tuple
            for element in arg:
                if isinstance(element, ir.Var) and not fld.input:
                    outputs.append(element)
        else:
            if isinstance(arg, ir.Var) and not fld.input:
                outputs.append(arg)
    return outputs

#--------------------------------------------------
# Filters
#--------------------------------------------------

def aggregate_outputs(agg: ir.Aggregate) -> list[ir.Var]:
    """ Get the list of vars bound to the outputs of this aggregation. """
    return list(filter(lambda arg: isinstance(arg, ir.Var) and not is_aggregate_input(arg, agg), agg.args)) # type: ignore

def aggregate_inputs(agg: ir.Aggregate) -> list[ir.Var]:
    """ Get the list of vars bound to the args that are inputs of this aggregation. """
    return list(filter(lambda arg: isinstance(arg, ir.Var) and is_aggregate_input(arg, agg), agg.args)) # type: ignore

def effective_logicals(tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Logical]:
    """ Filter tasks to return only the Logical tasks that are effective. """
    return OrderedSet.from_iterable(filter(lambda t: is_effective_logical(t), tasks))

def nullable_logicals(tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Logical]:
    """ Filter tasks to return only the Logical tasks that are nullable. """
    return OrderedSet.from_iterable(filter(lambda t: is_nullable_logical(t), tasks))

def hoisted_vars(hoisted: Iterable[ir.VarOrDefault]) -> list[ir.Var]:
    """ Extract the vars from defaults in the hoisted list, returning just Vars. """
    return [hoisted_var(v) for v in hoisted]

def hoisted_var(hoisted: ir.VarOrDefault) -> ir.Var:
    """ Extract the var from VarOrDefault, returning just Var. """
    return hoisted.var if isinstance(hoisted, ir.Default) else hoisted

def vars(args: Tuple[ir.Value, ...]) -> list[ir.Var]:
    """ Filter this list of values, keeping only Vars. """
    return cast(list[ir.Var], list(filter(lambda v: isinstance(v, ir.Var), flatten_tuple(args, ir.Value))))

def output_vars(aliases: FrozenOrderedSet[Tuple[str, ir.Value]]) -> list[ir.Var]:
    return [alias[1] for alias in aliases if isinstance(alias[1], ir.Var)]

def output_values(aliases: FrozenOrderedSet[Tuple[str, ir.Value]]) -> list[ir.Value]:
    return [alias[1] for alias in aliases]

def output_alias_names(aliases: FrozenOrderedSet[Tuple[str, ir.Value]]) -> list[str]:
    return [alias[0] for alias in aliases]

#--------------------------------------------------
# Visitors/Collectors
#--------------------------------------------------

def collect_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect all Vars starting at this node. """
    return cast(OrderedSet[ir.Var],
        visitor.collect_by_type(ir.Var, *nodes)
    )

def collect_quantified_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect all Vars that are children of Exists and ForAll. """
    return cast(OrderedSet[ir.Var],
        visitor.collect(
            lambda n, parent: isinstance(n, ir.Var) and isinstance(parent, (ir.Exists, ir.ForAll)),
            *nodes)
    )

def collect_aggregate_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars that are declared by aggregates in Rel (projection + over). """
    return cast(OrderedSet[ir.Var],
        # TODO - when dealing with multiple aggregations we will need to consider groupbys
        visitor.collect(
            lambda n, parent:
                # parent of var is an aggregate and either
                isinstance(parent, ir.Aggregate) and isinstance(n, ir.Var) and (
                # var is in the projection
                n in parent.projection or
                # var is an input to the aggregation's relation
                is_aggregate_input(n, parent)
                ),
            *nodes)
    )

def collect_rank_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars that are declared by ranks in Rel (projection + over). """
    return cast(OrderedSet[ir.Var],
        # TODO - when dealing with multiple aggregations we will need to consider groupbys
        visitor.collect(
            lambda n, parent:
                # parent of var is an aggregate and either
                isinstance(parent, ir.Rank) and isinstance(n, ir.Var) and (
                # var is in the projection
                n in parent.projection or
                n in parent.args
                ),
            *nodes)
    )

def collect_implicit_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars except the quantified and aggregate vars. """
    if not nodes:
        return ordered_set()
    return collect_vars(*nodes) - collect_quantified_vars(*nodes) - collect_aggregate_vars(*nodes) - collect_rank_vars(*nodes)

#--------------------------------------------------
# Useful node categories
#--------------------------------------------------

BINDERS = (ir.Lookup, ir.Construct, ir.Aggregate, ir.Exists, ir.Data, ir.Not)
COMPOSITES = (ir.Logical, ir.Sequence, ir.Union, ir.Match, ir.Until, ir.Wait)
EFFECTS = (ir.Update, ir.Output)

#--------------------------------------------------
# Helper classes
#--------------------------------------------------

class RewriteContext():
    """
    Container of information collected during a rewrite pass.
    """
    def __init__(self):
        # the logicals that will be at the top level at the end of the rewrite
        self.top_level: list[ir.Logical] = []
        # new relations created during the pass
        self.relations: list[ir.Relation] = []


#--------------------------------------------------
# Rewrite helpers
#--------------------------------------------------

def extract(task: ir.Task, body: OrderedSet[ir.Task], exposed_vars: list[ir.Var], ctx: RewriteContext, name: str) -> ir.Relation:
    """
    Extract into this Analysiscontext a new top level Logical that contains this body plus a
    derive task into a new temporary relation, which is also registered with the ctx.
    The exposed_vars determine the arguments of this temporary relation. The prefix
    can be used to customize the name of the relation, which defaults to the task kind.

    Return the temporary relation created for the extraction.
    """
    connection = create_connection_relation(task, exposed_vars, ctx, name)

    # add derivation to the extracted body
    body.add(f.derive(connection, exposed_vars))

    # extract the body
    ctx.top_level.append(clone_task(ir.Logical(task.engine, tuple(), tuple(body))))

    return connection

def create_connection_relation(task: ir.Task, exposed_vars: list[ir.Var], ctx: RewriteContext, name: str) -> ir.Relation:
    """
    Create a new relation with a name based off this task, with fields that represent
    the types and names of these exposed vars, and register in the context.
    """
    connection = f.relation(name, [f.field(v.name, v.type) for v in exposed_vars])
    ctx.relations.append(connection)

    return connection


def create_task_name(name_cache: NameCache, task: ir.Task, prefix: Optional[str]=None) -> str:
    """ Helper to generate consistent names for tasks extracted from a logical. """
    prefix = prefix if prefix else f"_{task.kind}"
    return name_cache.get_name(task.id, prefix)


CLONABLE = (ir.Var, ir.Default, ir.Task)
T = TypeVar('T', bound=ir.Task)
def clone_task(task: T) -> T:
    """
    Create a new task that is a clone of this task. This operation clones only sub-tasks
    and variables, and preserves variable references.

    This is useful when we are rewriting the metamodel and want to copy parts of a task to
    some other place. It is important to clone to avoid having the same object present
    multiple times in the metamodel.
    """

    # map from original object id to the rewritten object
    cache = {}
    def from_cache(original):
        """ Lookup this object from the cache above, dealing with collections and with
        objects that were not rewritten. """
        if isinstance(original, tuple):
            return tuple([from_cache(c) for c in original])
        elif isinstance(original, FrozenOrderedSet):
            return ordered_set(*[from_cache(c) for c in original])
        elif isinstance(original, CLONABLE):
            return cache.get(original.id, original)
        else:
            return original

    # the last node that was processed and rewritten
    prev_node = None
    stack: list[ir.Node] = [task]
    def to_stack(original):
        """ Add this original node to the stack if it is clonable and was never processed;
        return True iff the node was added to the stack. """
        if isinstance(original, CLONABLE) and original.id not in cache:
            stack.append(original)
            return True
        return False

    while stack:
        # peek the current node and get the initializable fields (i.e. ignore Node id)
        curr = stack[-1]
        curr_fields = list(filter(lambda f: f.init, fields(curr)))
        stacked_children = False

        # go over the fields adding to the stack the ones that we need to rewrite
        for field in curr_fields:
            field_value = getattr(curr, field.name)
            if isinstance(field_value, (tuple, FrozenOrderedSet)):
                # node field is a collection (tuple or set)
                for s in field_value:
                    if isinstance(s, tuple):
                        # the value within the collection is a tuple (can happen for lookup args)
                        for c in s:
                            stacked_children = to_stack(c) or stacked_children
                    else:
                        # the value within the collection is a scalar
                        stacked_children = to_stack(s) or stacked_children
            else:
                # node field is a scalar
                stacked_children = to_stack(field_value) or stacked_children

        # if no childrean were stacked, we rewrote all fields of curr, so we can pop it and rewrite it
        if not stacked_children:
            stack.pop()
            if curr.id not in cache:
                children = []
                for f in curr_fields:
                    children.append(from_cache(getattr(curr, f.name)))
                # create a new prev_node with the cloned children
                prev_node = curr.__class__(*children)
                cache[curr.id] = prev_node

    # the last node we processed is the rewritten original node
    assert(isinstance(prev_node, type(task)))
    return prev_node



import warnings

from v0.relationalai.semantics.lqp.constructors import (
    mk_abstraction, mk_and, mk_exists, mk_or, mk_pragma, mk_primitive,
    mk_specialized_value, mk_type, mk_value
)

__all__ = [
    "mk_abstraction", "mk_and", "mk_exists", "mk_or", "mk_pragma", "mk_primitive", "mk_specialized_value", "mk_type",
    "mk_value"
]

warnings.warn(
    "relationalai.early_access.lqp.constructors is deprecated. "
    "Please migrate to relationalai.semantics.lqp.constructors",
    DeprecationWarning,
    stacklevel=2,
)


# For builder components.
from v0.relationalai.semantics import Integer, define
from v0.relationalai.experimental.paths.graph import Graph


def ball_with_repetition(g:Graph, Source, max_length):
    edge = g.Edge
    Node = g.Node

    ball = g.model.Relationship(f"ball_rep {{Integer}} {{{Node}}}") # , [Integer, Node])

    u, v, src = Node.ref(), Node.ref(), Node.ref()
    level, m = Integer.ref(), Integer.ref()

    # Base case: a node is reachable from itself
    define(ball(0, src)).where(Source(src))

    # Recursive case:
    define(ball(level, v)).where(
        ball(m, u),
        m == level - 1,
        edge(u, v),
        level <= max_length
    )

    return ball



from __future__ import annotations
import time
from typing import Any, List, Optional, cast
from dataclasses import dataclass
import textwrap
from .. import dsl, std
from ..std import rel
from ..metamodel import Builtins
from ..tools.cli_controls import Spinner
from ..tools.constants import DEFAULT_QUERY_TIMEOUT_MINS
from .. import debugging
from .. errors import ResponseStatusException
import uuid
import v0.relationalai
import json
from ..clients.util import poll_with_specified_overhead
from ..clients.resources.snowflake import Resources as SnowflakeResources, APP_NAME
from ..clients.resources.snowflake import DirectAccessResources
from ..clients.direct_access_client import DirectAccessClient
from ..util.timeout import calc_remaining_timeout_minutes

rel_sv = rel._tagged(Builtins.SingleValued)

ENGINE_TYPE_SOLVER = "SOLVER"
# TODO (dba) The ERP still uses `worker` instead of `engine`. Change
# this once we fix this in the ERP.
WORKER_ERRORS = ["worker is suspended", "create/resume", "worker not found", "no workers found", "worker was deleted"]
ENGINE_ERRORS = ["engine is suspended", "create/resume", "engine not found", "no engines found", "engine was deleted"]
ENGINE_NOT_READY_MSGS = ["worker is in pending", "worker is provisioning", "worker is not ready to accept jobs"]

# --------------------------------------------------
# SolverModel object.
# --------------------------------------------------

class SolverModel:
    def __init__(self, graph:dsl.Graph):
        self.graph = graph
        self.id = dsl.next_id()
        self.scope = f"solvermodel{self.id}_"
        scope = self.scope
        self.Variable = dsl.Type(graph, "variables", scope=scope)
        self.MinObjective = dsl.Type(graph, "min_objectives", scope=scope)
        self.MaxObjective = dsl.Type(graph, "max_objectives", scope=scope)
        self.Constraint = dsl.Type(graph, "constraints", scope=scope)
        self.Solution = dsl.Type(graph, "solutions", scope=scope)
        self.components = [
            (self.MinObjective, "minimization objectives"),
            (self.MaxObjective, "maximization objectives"),
            (self.Constraint, "constraints"),
        ]
        self.solve_index = 0
        self.is_solved = False

        # Install model helpers.
        self.graph.install_raw(textwrap.dedent(f"""
            @inline
            def _solverlib_ho_appl(op, {{R}}, s): rel_primitive_solverlib_ho_appl(R, op, s)

            @inline
            def _solver_unwrap({{R}}, h, x...): exists((v) | R(v, x...) and pyrel_unwrap(v, h))

            declare {scope}variable_name
            declare {scope}component_name
            declare {scope}serialized
            declare {scope}primal_start

            def {scope}component_string(h, s):
                rel_primitive_solverlib_print_expr(
                    {scope}serialized[h], _solver_unwrap[{scope}variable_name], s
                )

            declare {scope}solve_output
        """))
        return None

    # Add an entity to the variable set, (optionally) set a string name from the
    # arguments and add domain constraints on the variable.
    def variable(
        self,
        var, # variable entity
        name_args:List|None=None, # list of strings to concatenate into a string name
        type:str|None=None, # variable type: "integer" or "zero_one"
        lower:int|float|None=None, # lower bound
        upper:int|float|None=None, # upper bound
        fixed:int|float|None=None, # fixed value
        start:int|float|None=None, # (primal) start value
    ):
        if type not in {"integer", "zero_one", None}:
            raise Exception(f"Invalid domain type: {type}.")
        var.set(self.Variable)

        # Set variable name.
        if name_args:
            var.set(**{f"{self.scope}variable_name": make_string(name_args)})

        # Add domain constraints.
        cons = []
        if fixed is not None:
            cons.append(eq(var, fixed))
        if type == "zero_one":
            cons.append(zero_one(var))
        if lower is not None and upper is not None:
            if type == "integer":
                cons.append(integer_interval(var, lower, upper))
            else:
                cons.append(interval(var, lower, upper))
        else:
            if type == "integer":
                cons.append(integer(var))
            if lower is not None:
                cons.append(gte(var, lower))
            if upper is not None:
                cons.append(lte(var, upper))
        if len(cons) == 1:
            self.constraint(cons[0])
        elif len(cons) > 1:
            self.constraint(and_(*cons))

        # Set primal start.
        if start is not None:
            var.set(**{f"{self.scope}primal_start": start})
        return var

    # Get variable string name.
    def variable_name(self, var):
        return std.alias(getattr(var, f"{self.scope}variable_name"), "name")

    # Add a constraint, minimization objective, or maximization objective.
    def constraint(self, expr, name_args:List|None=None):
        return self._add_component(self.Constraint, expr, name_args)

    def min_objective(self, expr, name_args:List|None=None):
        return self._add_component(self.MinObjective, expr, name_args)

    def max_objective(self, expr, name_args:List|None=None):
        return self._add_component(self.MaxObjective, expr, name_args)

    def _add_component(self, typ, expr, name_args:List|None):
        comp = typ.add(serialized=_wrap_expr(expr))
        if name_args:
            comp.set(component_name=make_string(name_args))
        return comp

    # Get component string name.
    def component_name(self, comp):
        return std.alias(comp.component_name, "name")

    # Get serialized component string in human-readable format.
    def component_string(self, comp):
        return std.alias(comp.component_string, "string")

    # Summarize the model by printing the number of variables and components.
    # Use outside a rule/query.
    def summarize(self):
        with self.graph.query() as select:
            vars = select(std.aggregates.count(self.Variable()))
        s = f"Model has: {vars.results.iat[0, 0]} variables"
        for (c_type, c_name) in self.components:
            with self.graph.query() as select:
                exprs = select(std.aggregates.count(c_type()))
            if not exprs.results.empty:
                s += f", {exprs.results.iat[0, 0]} {c_name}"
        print(s)
        return None

    # Print the model in human-readable format. Use outside a rule/query.
    def print(self):
        with self.graph.query() as select:
            vars = select(rel.last(getattr(rel, f"{self.scope}variable_name")))
        print("variables:")
        print(vars.results.to_string(index=False, header=False))
        for (c_type, c_name) in self.components:
            with self.graph.query() as select:
                exprs = select(self.component_string(c_type()))
            if not exprs.results.empty:
                print(c_name + ":")
                print(exprs.results.to_string(index=False, header=False))
        return None

    # Solve the model given a solver and solver options. Use outside a rule/query.
    def solve(self, solver: Solver, log_to_console=True, **kwargs):
        self.is_solved = False
        self.solve_index += 1
        self.solve_output = dsl.RelationNS([f"{self.scope}solve_output"], f"i_{self.solve_index}")

        options = kwargs
        options["version"] = 1

        # Validate options.
        for k, v in options.items():
            if not isinstance(k, str):
                raise Exception(f"Invalid parameter key. Expected string, got {type(k)} for {k}.")
            if not isinstance(v, (int, float, str, bool)):
                raise Exception(
                    f"Invalid parameter value. Expected string, integer, float, or boolean, got {type(v)} for {k}."
                )

        # Run the solve query and insert the solve_output result.
        scope = self.scope
        variable_name_string = f"{scope}variable_name" if "print_format" in options else "{}"
        component_name_string = f"{scope}component_name" if "print_format" in options else "{}"

        input_id = uuid.uuid4()
        model_uri = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-inputs/solver/{input_id}/model.binpb"
        sf_input_uri = f"snowflake://job-inputs/solver/{input_id}/model.binpb"

        payload: dict[str, Any] = {"solver": solver.solver_name.lower()}
        payload["options"] = options
        payload["model_uri"] = sf_input_uri

        rai_config = self.graph._config
        query_timeout_mins = kwargs.get("query_timeout_mins", None)
        if query_timeout_mins is None and (timeout_value := rai_config.get("query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS)) is not None:
            query_timeout_mins = int(timeout_value)
        config_file_path = getattr(rai_config, 'file_path', None)
        start_time = time.monotonic()
        remaining_timeout_minutes = query_timeout_mins
        response = self.graph.exec_raw(
            textwrap.dedent(f"""
            @inline
            def {scope}specialized_components(t, h, s):
                exists((v) | {{
                    (:min_objective, {scope}min_objectives);
                    (:max_objective, {scope}max_objectives);
                    (:constraint, {scope}constraints);
                    }}(t, v) and pyrel_unwrap(v, h) and {scope}serialized(v, s)
                )

            @no_diagnostics(:EXPERIMENTAL)
            def {scope}model_string {{
                rel_primitive_solverlib_model_string[{{
                    (:variable, _solver_unwrap[{scope}variables]);
                    {scope}specialized_components;
                    (:variable_name, _solver_unwrap[{variable_name_string}]);
                    (:expression_name, _solver_unwrap[{component_name_string}]);
                    (:primal_start, _solver_unwrap[{scope}primal_start]);
                }}]
            }}

            ic model_not_empty("Solver model is empty.") requires not empty({scope}model_string)

            def config[:envelope, :content_type]: "application/octet-stream"
            def config[:envelope, :payload, :data]: {scope}model_string
            def config[:envelope, :payload, :path]: "{model_uri}"
            def export {{ config }}
            """),
            query_timeout_mins=remaining_timeout_minutes,
        )
        txn = response.transaction or {}
        # The above `exec_raw` will throw an error if the transaction
        # gets aborted. But in the case it gets cancelled, by the user
        # or the system, it won't throw. In that case we also did not
        # upload the input model.
        if txn["state"] == "":
            txn = solver.provider.resources.get_transaction(txn["id"]) or {}
        if txn["state"] != "COMPLETED":
            raise Exception(f"Transaction that materializes the solver inputs did not complete! ID: `{txn['id']}` State `{txn['state']}`")

        # 2. Execute job and wait for completion.
        remaining_timeout_minutes = calc_remaining_timeout_minutes(
            start_time, query_timeout_mins, config_file_path=config_file_path
        )
        job_id = solver._exec_job(payload, log_to_console=log_to_console, query_timeout_mins=remaining_timeout_minutes)

        # 3. Extract result.
        remaining_timeout_minutes = calc_remaining_timeout_minutes(
            start_time, query_timeout_mins, config_file_path=config_file_path
        )
        res = self.graph.exec_raw(
            textwrap.dedent(f"""
            ic result_not_empty("Solver result is empty.") requires not empty(result)

            def result {{load_binary["snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-results/{job_id}/result.binpb"] }}
            def delete[:{scope}solve_output, :"i_{self.solve_index}"]: {scope}solve_output[:"i_{self.solve_index}"]

            @no_diagnostics(:EXPERIMENTAL)
            def insert[:{scope}solve_output, :"i_{self.solve_index}"]:
                rel_primitive_solverlib_extract[result]

            def output[:solver_error]: {scope}solve_output[:"i_{self.solve_index}", :error]
            """),
            readonly=False,
            query_timeout_mins=remaining_timeout_minutes,
        )
        errors = []
        for result in res.results:
            if result["relationId"] == "/:output/:solver_error/String":
                errors.extend(result["table"]["v1"])

        # 4. Map results to solution.
        with self.graph.rule(dynamic=True):
            sol = self.Solution.add(index = self.solve_index)
            for name in {"error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "printed_model"}:
                val = dsl.create_var()
                getattr(self.solve_output, name)(val)
                sol.set(**{name:val})

        self.is_solved = True

        if len(errors) > 0:
            raise Exception("\n".join(errors))

        return None

    # Get scalar result information after solving.
    def __getattr__(self, name:str):
        if not self.is_solved:
            raise Exception("Model has not been solved yet.")
        if name in {"error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "printed_model"}:
            return getattr(self.Solution(index = self.solve_index), name)
        else:
            return None

    # Get variable point values after solving. If `index` is specified, get the value
    # of the variable in the return the `index`-th solution.
    def value(self, var, index:int|None=None):
        if not self.is_solved:
            raise Exception("Model has not been solved yet.")
        val = dsl.create_var()
        unwrap_var = rel_sv.pyrel_unwrap(var)
        if index:
            self.solve_output.points(unwrap_var, index, val)
        else:
            self.solve_output.point(unwrap_var, val)
        std.alias(val, "value")
        return val

# --------------------------------------------------
# Operator definitions
# --------------------------------------------------

# Builtin binary operators

def plus(left, right):
    return _make_fo_expr(10, left, right)

def minus(left, right):
    return _make_fo_expr(11, left, right)

def mult(left, right):
    return _make_fo_expr(12, left, right)

def div(left, right):
    return _make_fo_expr(13, left, right)

def pow(left, right):
    return _make_fo_expr(14, left, right)

def eq(left, right):
    return _make_fo_expr(30, left, right)

def neq(left, right):
    return _make_fo_expr(31, left, right)

def lte(left, right):
    return _make_fo_expr(32, left, right)

def gte(left, right):
    return _make_fo_expr(33, left, right)

def lt(left, right):
    return _make_fo_expr(34, left, right)

def gt(left, right):
    return _make_fo_expr(35, left, right)

# First order operators

def abs(arg):
    return _make_fo_expr(20, arg)

def exp(arg):
    return _make_fo_expr(21, arg)

def log(arg):
    return _make_fo_expr(22, arg)

def integer(arg):
    return _make_fo_expr(41, arg)

def zero_one(arg):
    return _make_fo_expr(42, arg)

def interval(arg, low, high):
    return _make_fo_expr(51, low, high, arg)

def integer_interval(arg, low, high):
    return _make_fo_expr(50, low, high, 1, arg)

def if_then_else(cond, left, right):
    return _make_fo_expr(60, cond, left, right)

def not_(arg):
    return _make_fo_expr(61, arg)

def implies(left, right):
    return _make_fo_expr(62, left, right)

def iff(left, right):
    return _make_fo_expr(63, left, right)

def xor(left, right):
    return _make_fo_expr(64, left, right)

def and_(*args):
    return _make_fo_expr(70, *args)

def or_(*args):
    return _make_fo_expr(71, *args)

# Aggregate operators

def sum(*args, per=[]) -> Any:
    return _make_ho_expr(80, args, per)

def product(*args, per=[]) -> Any:
    return _make_ho_expr(81, args, per)

def min(*args, per=[]) -> Any:
    return _make_ho_expr(82, args, per)

def max(*args, per=[]) -> Any:
    return _make_ho_expr(83, args, per)

def count(*args, per=[]) -> Any:
    return _make_ho_expr(84, args, per)

def all_different(*args, per=[]) -> Any:
    return _make_ho_expr(90, args, per)

# --------------------------------------------------
# Symbolic expression helpers
# --------------------------------------------------

def _make_fo_expr(*args):
    expr = rel_sv.rel_primitive_solverlib_fo_appl(*args)
    expr.__class__ = SolverExpression
    return expr

# TODO(coey) test:
# dsl.tag(rel_sv.rel_primitive_solverlib_fo_appl, Builtins.Expensive)

def _make_ho_expr(op, args, per):
    return SolverExpression(dsl.get_graph(), _ho_appl_def, [args, per, [op]])

_ho_appl_def = dsl.build.aggregate_def("_solverlib_ho_appl")

class SolverExpression(dsl.Expression):
    def __init__(self, graph, op, args):
        super().__init__(graph, op, args)

def _wrap_expr(e):
    # If expression is not known to produce a serialized expression string,
    # wrap it with the identity operation just in case
    return e if isinstance(e, SolverExpression) else _make_fo_expr(0, e)

# Symbolic expression context, in which some builtin infix operators are redefined
# TODO(coey) handle comparison chains (e.g. 0 < x < y <= 1) or throw error
class Operators(dsl.Context):
    def _supports_binary_op(self, op):
        return op in _builtin_binary_map

    def _make_binary_op(self, op, left, right):
        return _make_fo_expr(_builtin_binary_map[op], left, right)

def operators():
    return Operators(dsl.get_graph())

# Maps for Builtins operator to SolverLib operator ID
_builtin_binary_map = {
    Builtins.plus: 10,
    Builtins.minus: 11,
    Builtins.mult: 12,
    Builtins.div: 13,
    Builtins.pow: 14,
    Builtins.approx_eq: 30,
    Builtins.neq: 31,
    Builtins.lte: 32,
    Builtins.gte: 33,
    Builtins.lt: 34,
    Builtins.gt: 35,
}

# Concatenate arguments into a string separated by underscores
def make_string(args:List):
    string = args[0]
    for arg in args[1:]:
        string = rel_sv.concat(rel_sv.concat(string, "_"), arg)
    return string


# --------------------------------------------------
# Solver
# --------------------------------------------------


@dataclass
class PollingState:
    job_id: str
    continuation_token: str
    is_done: bool
    log_to_console: bool


class Solver:
    def __init__(
        self,
        solver_name: str,
        engine_name: str | None = None,
        engine_size: str | None = None,
        auto_suspend_mins: int | None = None,
        resources: SnowflakeResources | None = None,
    ):
        self.provider = Provider(resources=resources)
        self.solver_name = solver_name.lower()

        self.rai_config = self.provider.resources.config
        settings: dict[str, Any] = {}
        if "experimental" in self.rai_config:
            exp_config = self.rai_config.get("experimental", {})
            if isinstance(exp_config, dict):
                if "solvers" in exp_config:
                    settings = exp_config["solvers"].copy()

        # Engine configuration fields are not necessary for the solver
        # settings so we `pop` them from the settings object. Default
        # size and auto_suspend_mins are set in the `Provider` methods.
        engine_name = engine_name or settings.pop("engine", None)
        if not engine_name:
            engine_name = self.provider.resources.get_user_based_engine_name()
        self.engine_name = engine_name

        self.engine_size = engine_size or settings.pop("engine_size", None)
        self.engine_auto_suspend_mins = auto_suspend_mins or settings.pop("auto_suspend_mins", None)

        # The settings are used when creating a solver engine, they
        # may configure each individual solver.
        self.engine_settings = settings

        # Optimistically set the engine object to a `READY` engine to
        # avoid checking the engine status on each execution.
        self.engine:Optional[dict[str,Any]] = {"name": engine_name, "state": "READY"}

        return None

    # --------------------------------------------------
    # Helper
    # --------------------------------------------------
    def _auto_create_solver_async(self):
        name = self.engine_name
        auto_suspend_mins = self.engine_auto_suspend_mins
        size = self.engine_size
        settings = self.engine_settings
        with Spinner(
            "Checking solver status",
            leading_newline=True,
        ) as spinner:
            engine = None
            engines = [e for e in self.provider.list_solvers() if e["name"] == name]
            assert len(engines) == 1 or len(engines) == 0
            if len(engines) != 0:
                engine = engines[0]

            if engine:
                # TODO (dba) Logic engines support altering the
                # auto_suspend_mins setting. Currently, we don't have
                # this capability for solver engines, so users need to
                # recreate or use another engine. For both the size
                # and Gurobi configuration the user anyways has to
                # create a new one as this configuration must happen
                # when the engine is created.
                settings_cannot_be_altered_msg = "The configuration of a solver engine happens when the engine is created and _cannot_ be changed. You either need to specify a new engine name or delete the current engine.\n\nSee `solvers.Provider().delete_solver()`."
                # Make sure that the solver requested is enabled
                # on the engine.
                if self.solver_name not in engine["solvers"]:
                    raise Exception(
                        f"Solver `{self.solver_name}` is not enabled on `{name}`.\n\n" + settings_cannot_be_altered_msg
                    )

                # Make sure size and auto_suspend_mins settings match
                # what the user requests.
                if size is not None and size != engine["size"]:
                    raise Exception(
                        f"Engine `{name}` has size setting of `{engine['size']}` but size is requested to be `{size}`.\n\n" + settings_cannot_be_altered_msg
                    )

                if auto_suspend_mins is not None and auto_suspend_mins != engine["auto_suspend_mins"]:
                    raise Exception(
                        f"Engine `{name}` has auto_suspend_mins setting of `{engine['auto_suspend_mins']}` but auto_suspend_mins is requested to be `{auto_suspend_mins}`.\n\n" + settings_cannot_be_altered_msg
                    )

                if engine["state"] == "PENDING":
                    spinner.update_messages(
                        {
                            "finished_message": f"Solver {name} is starting",
                        }
                    )
                    pass
                elif engine["state"] == "SUSPENDED":
                    spinner.update_messages(
                        {
                            "finished_message": f"Resuming solver {name}",
                        }
                    )
                    self.provider.resume_solver_async(name)
                elif engine["state"] == "READY":
                    spinner.update_messages(
                        {
                            "finished_message": f"Solver {name} is ready",
                        }
                    )
                    pass
                else:
                    spinner.update_messages(
                        {
                            "message": f"Restarting solver {name}",
                        }
                    )
                    self.provider.delete_solver(name)
                    engine = None
            if not engine:
                # Validate Gurobi config.
                if self.solver_name == "gurobi":
                    is_gurobi_configured = False
                    gurobi_config = settings.get("gurobi", {})
                    if all(
                        k in gurobi_config
                        for k in ["license_secret_name", "external_access_integration"]
                    ):
                        is_gurobi_configured = True
                    if not is_gurobi_configured:
                        raise Exception(
                            "Gurobi is not properly configured. You need to provide both `license_secret_name` and `external_access_integration` in its configuration, see https://docs.relational.ai/build/reasoners/prescriptive/solver-backends/gurobi/#usage"
                        )
                self.provider.create_solver_async(name, settings=settings, size=size, auto_suspend_mins=auto_suspend_mins)
                engine = self.provider.get_solver(name)
                spinner.update_messages(
                    {
                        "finished_message": f"Starting solver {name}...",
                    }
                )

            self.engine = engine

    def _exec_job(self, payload, log_to_console=True, query_timeout_mins: Optional[int]=None):
        if self.engine is None:
            raise Exception("Engine not initialized.")

        with debugging.span("job") as job_span:
            # Retry logic. If creating a job fails with an engine
            # related error we will create/resume/... the engine and
            # retry.
            try:
                job_id = self.provider.create_job_async(self.engine["name"], payload, query_timeout_mins=query_timeout_mins)
            except Exception as e:
                err_message = str(e).lower()
                if isinstance(e, ResponseStatusException):
                    err_message = e.response.json().get("message", "")
                if any(kw in err_message.lower() for kw in ENGINE_ERRORS + WORKER_ERRORS + ENGINE_NOT_READY_MSGS):
                    self._auto_create_solver_async()
                    # Wait until the engine is ready.
                    poll_with_specified_overhead(lambda: self._is_solver_ready(), 0.1)
                    job_id = self.provider.create_job_async(self.engine["name"], payload, query_timeout_mins=query_timeout_mins)
                else:
                    raise e

            job_span["job_id"] = job_id
            debugging.event("job_created", job_span, job_id=job_id, engine_name=self.engine["name"], job_type=ENGINE_TYPE_SOLVER)
            if not isinstance(job_id, str):
                job_id = ""
            polling_state = PollingState(job_id, "", False, log_to_console)

            try:
                with debugging.span("wait", job_id=job_id):
                    poll_with_specified_overhead(
                        lambda: self._check_job_status(polling_state), 0.1
                    )
            except KeyboardInterrupt as e:
                print(f"Canceling job {job_id}")
                self.provider.cancel_job(job_id)
                raise e

            return job_id

    def _is_solver_ready(self):
        if self.engine is None:
            raise Exception("Engine not initialized.")

        result = self.provider.get_solver(self.engine["name"])

        if result is None:
            raise Exception("No engine available.")
            
        self.engine = result
        state = result["state"]
        if state != "READY" and state != "PENDING":
            # Might have suspended or otherwise gone. Recreate.
            self._auto_create_solver_async()
        return state == "READY"

    def _check_job_status(self, state):
        response = self.provider.get_job(state.job_id)
        assert response, f"No results from get_job('{state.job_id}')"

        status: str = response["state"]

        self._print_solver_logs(state)

        return status == "COMPLETED" or status == "FAILED" or status == "CANCELED"

    def _print_solver_logs(self, state: PollingState):
        if state.is_done:
            return

        resp = self.provider.get_job_events(state.job_id, state.continuation_token)

        # Print solver logs to stdout.
        for event in resp["events"]:
            if event["type"] == "LogMessage":
                if state.log_to_console:
                    print(event["event"]["message"])
            else:
                continue

        state.continuation_token = resp["continuation_token"]
        if state.continuation_token == "":
            state.is_done = True


# --------------------------------------------------
# Provider
# --------------------------------------------------
#
# TODO (dba) We use an experimental and unified engine API for
# solvers. Once it is no longer experimental we can remove the
# provider here and use the normal PyRel provider.


class Provider:
    def __init__(self, resources=None):
        if not resources:
            resources = v0.relationalai.Resources()
        if not isinstance(resources, SnowflakeResources):
            raise Exception("Solvers are only supported on SPCS.")

        # Type narrowing: resources is confirmed to be SnowflakeResources
        self.resources: SnowflakeResources = cast(SnowflakeResources, resources)
        self.direct_access_client: Optional[DirectAccessClient] = None

        if isinstance(self.resources, DirectAccessResources):
            self.direct_access_client = self.resources.direct_access_client

    def create_solver(
        self,
        name: str,
        size: str | None = None,
        settings: dict | None = None,
        auto_suspend_mins: int | None = None,
    ):
        if size is None:
            size = "HIGHMEM_X64_S"
        if settings is None:
            settings = {}
        engine_config: dict[str, Any] = {"settings": settings}
        if auto_suspend_mins is not None:
            engine_config["auto_suspend_mins"] = auto_suspend_mins
        self.resources._exec_sql(
            f"CALL {APP_NAME}.experimental.create_engine('{ENGINE_TYPE_SOLVER}', '{name}', '{size}', {engine_config});", None
        )

    def create_solver_async(
        self,
        name: str,
        size: str | None = None,
        settings: dict | None = None,
        auto_suspend_mins: int | None = None,
    ):
        if size is None:
            size = "HIGHMEM_X64_S"

        if self.direct_access_client is not None:
            payload:dict[str, Any] = {
                "name": name,
                "settings": settings,
               }
            if auto_suspend_mins is not None:
                payload["auto_suspend_mins"] = auto_suspend_mins
            if size is not None:
                payload["size"] = size
            response = self.direct_access_client.request(
                "create_engine",
                payload=payload,
                path_params={"engine_type": "solver"},
               )
            if response.status_code != 200:
                raise ResponseStatusException(
                    f"Failed to create engine {name} with size {size}.", response
                )
        else:
            engine_config: dict[str, Any] = {}
            if settings is not None:
                engine_config["settings"] = settings
            if auto_suspend_mins is not None:
                engine_config["auto_suspend_mins"] = auto_suspend_mins
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.create_engine_async('{ENGINE_TYPE_SOLVER}', '{name}', '{size}', {engine_config});",
                None
               )

    def delete_solver(self, name: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "delete_engine", path_params = {"engine_type": ENGINE_TYPE_SOLVER, "engine_name": name}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to delete engine.", response)
            return None
        else:
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.delete_engine('{ENGINE_TYPE_SOLVER}', '{name}');",
                None
            )

    def resume_solver_async(self, name: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "resume_engine", path_params = {"engine_type": ENGINE_TYPE_SOLVER, "engine_name": name}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to resume engine.", response)
            return None
        else:
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.resume_engine_async('{ENGINE_TYPE_SOLVER}', '{name}');",
                None
            )
            return None

    def get_solver(self, name: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "get_engine", path_params = {"engine_type": ENGINE_TYPE_SOLVER, "engine_name": name}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to get engine.", response)
            solver = response.json()
            if not solver :
                return None
            solver_state = {
                "name": solver["name"],
                "id": solver["id"],
                "size": solver["size"],
                "state": solver["status"], # callers are expecting 'state'
                "created_by": solver["created_by"],
                "created_on": solver["created_on"],
                "updated_on": solver["updated_on"],
                "version": solver["version"],
                "auto_suspend": solver["auto_suspend_mins"],
                "suspends_at": solver["suspends_at"],
                "solvers": []
                if solver["settings"] == ""
                else [
                        k
                        for (k,v) in json.loads(solver["settings"]).items()
                        if isinstance(v, dict) and v.get("enabled", False)
                ],
                }
            return solver_state
        else:
            results = self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.get_engine('{ENGINE_TYPE_SOLVER}', '{name}');",
                None
               )
            return solver_list_to_dicts(results)[0]

    def list_solvers(self, state: str | None = None):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "list_engines"
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to list engines.", response)
            response_content = response.json()
            if not response_content:
                return []
            engines = [
                {
                    "name": engine["name"],
                    "id": engine["id"],
                    "size": engine["size"],
                    "state": engine["status"], # callers are expecting 'state'
                    "created_by": engine["created_by"],
                    "created_on": engine["created_on"],
                    "updated_on": engine["updated_on"],
                    "auto_suspend_mins": engine["auto_suspend_mins"],
                    "solvers": []
                    if engine["settings"] == ""
                    else [
                        k
                        for (k, v) in json.loads(engine["settings"]).items()
                        if isinstance(v, dict) and v.get("enabled", False)
                           ],
                }
                for engine in response_content.get("engines", [])
                if (state is None or engine.get("status") == state) and (engine.get("type") == ENGINE_TYPE_SOLVER)
               ]
            return sorted(engines, key=lambda x: x["name"])
        else:
            where_clause = f"WHERE TYPE='{ENGINE_TYPE_SOLVER}'"
            where_clause = (
                f"{where_clause} AND STATUS = '{state.upper()}'" if state else where_clause
               )
            statement = f"SELECT NAME,ID,SIZE,STATUS,CREATED_BY,CREATED_ON,UPDATED_ON,AUTO_SUSPEND_MINS,SETTINGS FROM {APP_NAME}.experimental.engines {where_clause};"
            results = self.resources._exec_sql(statement, None)
            return solver_list_to_dicts(results)

    # --------------------------------------------------
    # Job API
    # --------------------------------------------------

    def create_job_async(self, engine_name, payload, query_timeout_mins: Optional[int]=None):
        payload_json = json.dumps(payload)

        if query_timeout_mins is None and (timeout_value := self.resources.config.get("query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS)) is not None:
            query_timeout_mins = int(timeout_value)
            
        if self.direct_access_client is not None:
            job = {
                "job_type":ENGINE_TYPE_SOLVER,
                "worker_name": engine_name,
                "timeout_mins": query_timeout_mins,
                "payload": payload_json,
            }
            response = self.direct_access_client.request(
                "create_job",
                payload=job,
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to create job.", response)
            response_content = response.json()
            return response_content["id"]
        else:
            if query_timeout_mins is not None:
                sql_string = textwrap.dedent(f"""
                CALL {APP_NAME}.experimental.exec_job_async('{ENGINE_TYPE_SOLVER}', '{engine_name}', '{payload_json}', null, {query_timeout_mins})
                """)
            else:
                sql_string = textwrap.dedent(f"""
                CALL {APP_NAME}.experimental.exec_job_async('{ENGINE_TYPE_SOLVER}', '{engine_name}', '{payload_json}')
                """)
            res = self.resources._exec_sql(sql_string, None)
            return res[0]["ID"]

    def list_jobs(self, state=None, limit=None):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "list_jobs"
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to list jobs.", response)
            response_content = response.json()
            if not response_content:
                return []
            jobs = [
                {
                    "id": job["id"],
                    "state": job["state"],
                    "created_by": job["created_by"],
                    "created_on": job["created_on"],
                    "finished_at": job.get("finished_at", None),
                    "duration": job["duration"] if "duration" in job else 0,
                    "solver": json.loads(job["payload"]).get("solver", ""),
                    "engine": job.get("engine_name", job["worker_name"]),
                }
                for job in response_content.get("jobs", [])
                if state is None or job.get("state") == state
               ]
            return sorted(jobs, key=lambda x: x["created_on"], reverse=True)
        else:
            state_clause = f"AND STATE = '{state.upper()}'" if state else ""
            limit_clause = f"LIMIT {limit}" if limit else ""
            results = self.resources._exec_sql(
                f"SELECT ID,STATE,CREATED_BY,CREATED_ON,FINISHED_AT,DURATION,PAYLOAD,ENGINE_NAME FROM {APP_NAME}.experimental.jobs where type='{ENGINE_TYPE_SOLVER}' {state_clause} ORDER BY created_on DESC {limit_clause};",
                None
               )
            return job_list_to_dicts(results)

    def get_job(self, id: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "get_job", path_params = {"job_type": ENGINE_TYPE_SOLVER, "job_id": id}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to get job.", response)
            response_content = response.json()
            return response_content["job"]
        else:
            results = self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.get_job('{ENGINE_TYPE_SOLVER}', '{id}');",
                None
            )
            return job_list_to_dicts(results)[0]

    def get_job_events(self, job_id: str, continuation_token: str = ""):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "get_job_events",
                path_params = {"job_type": ENGINE_TYPE_SOLVER, "job_id": job_id, "stream_name": "progress"},
                query_params={"continuation_token": continuation_token},
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to get job events.", response)
            response_content = response.json()
            if not response_content:
                return {
                    "events": [],
                    "continuation_token": None
                }
            return response_content
        else:
            results = self.resources._exec_sql(
                f"SELECT {APP_NAME}.experimental.get_job_events('{ENGINE_TYPE_SOLVER}', '{job_id}', '{continuation_token}');",
                None
               )
            if not results:
                return {"events": [], "continuation_token": None}
            row = results[0][0]
            if not isinstance(row, str):
                row = ""
            return json.loads(row)

    def cancel_job(self, id: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "cancel_job", path_params = {"job_type": ENGINE_TYPE_SOLVER, "job_id": id}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to cancel job.", response)
            return None
        else:
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.cancel_job('{ENGINE_TYPE_SOLVER}', '{id}');",
                None
            )
            return None


def solver_list_to_dicts(results):
    if not results:
        return []
    return [
        {
            "name": row["NAME"],
            "id": row["ID"],
            "size": row["SIZE"],
            "state": row["STATUS"],  # callers are expecting 'state'
            "created_by": row["CREATED_BY"],
            "created_on": row["CREATED_ON"],
            "updated_on": row["UPDATED_ON"],
            "auto_suspend_mins": row["AUTO_SUSPEND_MINS"],
            "solvers": []
            if row["SETTINGS"] == ""
            else [
                k
                for (k, v) in json.loads(row["SETTINGS"]).items()
                if isinstance(v, dict) and v.get("enabled", False)
            ],
        }
        for row in results
    ]

def job_list_to_dicts(results):
    if not results:
        return []
    return [
        {
            "id": row["ID"],
            "state": row["STATE"],
            "created_by": row["CREATED_BY"],
            "created_on": row["CREATED_ON"],
            "finished_at": row["FINISHED_AT"],
            "duration": row["DURATION"] if "DURATION" in row else 0,
            "solver": json.loads(row["PAYLOAD"])["solver"],
            "engine": row["ENGINE_NAME"],
        }
        for row in results
    ]



from abc import ABC, abstractmethod
from typing import Union, Optional

from v0.relationalai.early_access.dsl.orm.measures.initializer import get_reasoner
from v0.relationalai.early_access.dsl.orm.models import Role
from v0.relationalai.semantics.metamodel.util import OrderedSet, ordered_set

class RoleExpr(ABC):

    def __init__(self):
        self._reasoner = get_reasoner()

    @staticmethod
    def flat_concat(x1, x2):
        def flatten(expr):
            return expr._seq if expr.composer() else [expr]

        parts = flatten(x1) + flatten(x2)
        return ComposeRoleExpr(*parts)

    def ending_roles(self) -> OrderedSet[Role] : return ordered_set()

    @abstractmethod
    def guid(self) -> int: pass

    def composer(self) -> bool: return False
    def nil(self) -> bool: return False
    def kleene_plus(self) -> bool: return False
    def kleene_star(self) -> bool: return False
    def simple(self) -> bool: return False
    def mandatory(self) -> bool: return False

    def plus(self):
        return PlusRoleExpr(self)

    def star(self):
        return StarRoleExpr(self)

    def subsumes(self, other) -> Union[tuple['RoleExpr', Optional['RoleExpr']], tuple[None, None]]:
        return (self, None) if other.nil() else (None, None)

    @abstractmethod
    def pretty_print(self) -> str:
        pass

class NilRoleExpr(RoleExpr):

    def __init__(self):
        super().__init__()

    def guid(self) -> int: return hash("Nil")

    def mandatory(self): return True

    def nil(self): return True

    def pretty_print(self) -> str: return "Nil"

class SimpleRoleExpr(RoleExpr):

    def __init__(self, r: Role):
        super().__init__()
        if not isinstance(r, Role):
            raise Exception(f"Tried to instantiate a SimpleRoleExpr with a non-role: {r} of type {type(r)}")
        self._role = r
        
    def ending_roles(self): return ordered_set(self._role)

    def guid(self) -> int: return hash(self._role._guid())

    def simple(self): return True

    def mandatory(self):
        this = self._role

        if not self._reasoner.is_one_role(this):
            return False

        roles = this.siblings()
        if len(roles) > 1:
            return False

        return self._reasoner.is_mandatory_role(roles[0])

    def pretty_print(self) -> str:
        return f"{self._role}"

    def subsumes(self, other):
        if other.simple():
            return (self, None) if self._role._guid() == other._role._guid() else (None, None)
        return super().subsumes(other)

class UnionRoleExpr(RoleExpr):
    def __init__(self, *roles):
        super().__init__()

        if not roles:
            raise ValueError("At least one role must be provided for the UnionRoleExpr.")

        self._roles = roles

        matcher=None
        for r in roles:
            if r is None:
                raise ValueError("NoneType role provided to UnionRoleExpr.")
            if matcher is None:
                matcher = r
            else:
                if not self._reasoner.are_roles_compatible(r, matcher):
                    raise Exception(f"Two roles {r} and {matcher} with incompatible players used in a union role expression")

        assert matcher is not None, "Matcher should not be None after processing roles."
        self._role_player = matcher.player()

    def common_player(self): return self._role_player

    def guid(self) -> int:
        role_hashes = [hash(r) for r in self._roles]
        return hash(tuple(role_hashes))

    def pretty_print(self) -> str:
        expr = " | ".join([f"{r}" for r in self._roles])
        return f"( {expr} )"

    def roles(self): return self._roles

class ComposeRoleExpr(RoleExpr):

    def __init__(self, *exprs):
        super().__init__()
        self._seq = []
        for exp in exprs:
            if not isinstance(exp, RoleExpr):
                raise Exception(f"Tried to concatenate with non- role expr: {exp} of type {type(exp)}")
            self._seq.append(exp)

    def composer(self) -> bool: return True

    # Finds all ending roles of this ConcatExpr. We have to iterate
    # in reverse until we find the first non Kleene star expr to make
    # sure we account for exprs like:
    #    r1 . r2 . ( r3 )* . (r4 . r5)*
    # which should return { r2, r3, r5 }
    #
    def ending_roles(self):
        roles = ordered_set()
        for expr in reversed(self._seq):
            roles.update(expr.ending_roles())
            if not expr.kleene_star():
                break
        return roles

    def guid(self) -> int:
        role_hashes = [x.guid() for x in self._seq]
        return hash(tuple(role_hashes))

    def mandatory(self):
        for e in self._seq:
            if not e.mandatory():
                return False
        return True

    def pretty_print(self) -> str:
        parts = [ p.pretty_print() for p in self._seq ]
        return " . ".join(parts)

    def subsumes(self, other):

        if other.nil():
            return self, None

        o_seq = other._seq if other.composer() else [other]
        my_seq = self._seq
        o_seq_length = len(o_seq)

        if len(my_seq) < o_seq_length:
            return None, None

        splice_point_stack = []
        for i in range(o_seq_length):
            my_expr = my_seq[i]
            (x, splicer) = my_expr.subsumes(o_seq[i])
            if x is None:
                return None, None
            else:
                if my_expr.kleene_star():
                    splice_point_stack.append(my_expr)
                else:
                    splice_point_stack = [my_expr]

        # Upon exit, splice_point_stack must contain at least one role expr,
        # and because len(o_seq) could be less than len(my_seq) we need to
        # consider any remaining exprs in my_seq when creating the splicer.

        if len(splice_point_stack) == 1:
            join_expr = splice_point_stack[0]
        else:
            oset = ordered_set()
            for x in splice_point_stack:
                for r in x.ending_roles():
                    oset.add(r)
            join_expr = UnionRoleExpr(*oset)


        if len(my_seq) == o_seq_length:
            return join_expr, None
        else:
            if len(my_seq) - o_seq_length == 1:
                return join_expr, my_seq[o_seq_length]
            else:
                # Create a composer that represents what to splice onto
                # any dimension at the join point denoted by join_expr
                expr_seq = []
                for i in range(o_seq_length, len(my_seq), 1):
                    expr_seq.append(my_seq[i])

            return join_expr, ComposeRoleExpr(*expr_seq)

class ClosureRoleExpr(RoleExpr):
    def __init__(self, x):
        super().__init__()
        if not isinstance(x, RoleExpr):
            raise Exception(f"Tried to take the reflexive a closure of something other than a role expression: {x} of type {type(x)}")
        self._component = x

    def ending_roles(self):
        return self._component.ending_roles()

    def mandatory(self):
        return self._component.mandatory()
    
    def role_body(self):
        assert(isinstance(self._component, SimpleRoleExpr))
        return self._component._role

class StarRoleExpr(ClosureRoleExpr):
    def __init__(self, x):
        super().__init__(x)

    def guid(self) -> int: return hash(("*", self._component.guid()))

    def kleene_star(self) -> bool: return True

    def pretty_print(self) -> str:
        return f"( {self._component.pretty_print()} ) *"

    def subsumes(self, other):
        if other.kleene_star():
            (x, _) = self._component.subsumes(other._component)
            return (x, None) if x is not None else (None, None)
        return super().subsumes(other)

class PlusRoleExpr(ClosureRoleExpr):
    def __init__(self, x):
        super().__init__(x)

    def guid(self) -> int: return hash(("+", self._component.guid()))

    def kleene_plus(self) -> bool: return True

    def pretty_print(self) -> str:
        return f"( {self._component.pretty_print()} ) +"

    def subsumes(self, other):
        if other.kleene_plus():
            (x, _) = self._component.subsumes(other._component)
            return (x, None) if x is not None else (None, None)
        return super().subsumes(other)


from __future__ import annotations

from v0.relationalai.semantics.internal import internal as i
from v0.relationalai.semantics.metamodel.util import OrderedSet
from .std import _Integer, _String, _make_expr
from typing import Literal, Any
from .. import std


def escape(regex: _String) -> i.Expression:
    return _make_expr("escape_regex_metachars", regex, i.String.ref())

class Match(i.Producer):

    def __init__(self, regex: _String, string: _String, pos: _Integer = 0, _type: Literal["search", "fullmatch", "match"] = "match"):
        super().__init__(i.find_model([regex, string, pos]))
        self.regex = regex
        self.string = string
        self.pos = pos

        if _type == "match":
            self._expr = _regex_match_all(self.regex, self.string, std.cast_to_int64(self.pos + 1))
            self._offset, self._full_match = self._expr._arg_ref(2), self._expr._arg_ref(3)
        elif _type == "search":
            raise NotImplementedError("`search` is not implemented")
        elif _type == "fullmatch":
            _exp = _regex_match_all(self.regex, self.string, std.cast_to_int64(self.pos + 1))
            self._offset, self._full_match = _exp._arg_ref(2), _exp._arg_ref(3)
            self._expr = self._full_match == std.strings.substring(self.string, std.cast_to_int64(self.pos), std.strings.len(self.string))

    def group(self, index: _Integer = 0) -> i.Producer:
        if index == 0:
            return self._full_match
        else:
            return _make_expr("capture_group_by_index", self.regex, self.string, std.cast_to_int64(self.pos + 1), std.cast_to_int64(index), i.String.ref("res"))

    def group_by_name(self, name: _String) -> i.Producer:
        return _make_expr("capture_group_by_name", self.regex, self.string, std.cast_to_int64(self.pos + 1), name, i.String.ref("res"))

    def start(self) -> i.Expression:
        return self._offset - 1

    def end(self) -> i.Expression:
        return std.strings.len(self.group(0)) + self.start() - 1

    def span(self) -> tuple[i.Producer, i.Producer]:
        return self.start(), self.end()

    def _to_keys(self) -> OrderedSet[Any]:
        return i.find_keys(self._expr)

    def _compile_lookup(self, compiler:i.Compiler, ctx:i.CompilerContext):
        compiler.lookup(self.regex, ctx)
        compiler.lookup(self.string, ctx)
        compiler.lookup(self.pos, ctx)
        return compiler.lookup(self._expr, ctx)

    def __getattr__(self, name: str) -> Any:
        return object.__getattribute__(self, name)

    def __setattr__(self, name: str, value: Any) -> None:
        object.__setattr__(self, name, value)


def match(regex: _String, string: _String) -> Match:
    return Match(regex, string)

def search(regex: _String, string: _String, pos: _Integer = 0) -> Match:
    return Match(regex, string, pos, _type="search")

def fullmatch(regex: _String, string: _String, pos: _Integer = 0) -> Match:
    return Match(regex, string, pos, _type="fullmatch")

def findall(regex: _String, string: _String) -> tuple[i.Producer, i.Producer]:
    exp = _regex_match_all(regex, string)
    ix, match = exp._arg_ref(2), exp._arg_ref(3)
    rank = i.rank(i.asc(ix, match))
    return rank, match

def _regex_match_all(regex: _String, string: _String, pos: _Integer|None = None) -> i.Expression:
    if pos is None:
        pos = i.Int64.ref()
    return _make_expr("regex_match_all", regex, string, pos, i.String.ref())


# In your_package/submodule.py
from .internal import Relationship

external = Relationship.builtins["external"]
concept_population = Relationship.builtins["concept_population"]
function = Relationship.builtins["function"]
from_cdc = Relationship.builtins["from_cdc"]
track = Relationship.builtins["track"]
recursion_config = Relationship.builtins["recursion_config"]



import keyword
from typing import Any, Callable, Dict, Type

from relationalai.util.structures import OrderedSet
from .metamodel import (
    Node, Model, NumberType, ScalarType, Task, Logical, Sequence, Union, Match, Until, Wait, Loop, Break, Require,
    Not, Exists, Lookup, Update, Aggregate, Construct, Var, Annotation, Field, Literal, UnionType, ListType, TupleType
)
import datetime as _dt

#------------------------------------------------------
# Constants
#------------------------------------------------------

_ATOMS = (str, int, float, bool, type(None), _dt.datetime)
NO_WALK = object()

WALK_FIELDS: Dict[Type[Node], tuple[str, ...]] = {
    # Top-level
    Model: ("root",),

    # Control flow
    Logical: ("body",),
    Sequence: ("tasks",),
    Union: ("tasks",),
    Match: ("tasks",),
    Until: ("check", "body",),
    Wait: ("check",),
    Loop: ("over", "body",),
    Break: ("check",),

    # Constraints / Quantifiers
    Require: ("domain", "check", "error"),
    Not: ("task",),
    Exists: ("vars", "task"),

    # Relation ops
    Lookup: ("args",),  # relation has no tasks/vars; args is ISeq[Value]
    Update: ("args",),
    Aggregate: ("projection", "group", "args", "body"),
    Construct: ("values", "id_var"),

    # Annotations can contain Values which can contain Vars
    Annotation: ("args",),

    # Var: no children, but we include it so enter_var/var() can run
    Var: (),
    Field: (),
    Literal: (),
    ScalarType: (),
    NumberType: (),
    UnionType: (),
    ListType: (),
    TupleType: (),
}

#------------------------------------------------------
# Walker
#------------------------------------------------------

class Walker:
    visits = 0
    _walkers: Dict[Type[Node], Callable] = {}
    __slots__ = ("_seen_ids",)

    def __init__(self):
        # ids of nodes on the current recursion path (for cycle detection)
        self._seen_ids: set[int] = set()

    def __call__(self, obj):
        self._seen_ids.clear()
        try:
            self._walk(obj)
        except StopIteration:
            pass
        return obj

    def _walk(self, x):
        if isinstance(x, Node):
            return self._visit_node(x)
        if isinstance(x, _ATOMS):
            return x
        if isinstance(x, (tuple, list, OrderedSet)):
            for v in x:
                self._walk(v)
            return x
        if isinstance(x, dict):
            for v in x.values():
                self._walk(v)
            return x
        return x

    def _visit_node(self, node: Node):
        nid = id(node)
        if nid in self._seen_ids:  # cycle guard by identity
            return node

        Walker.visits += 1
        self._seen_ids.add(nid)

        cls = type(node)
        self._walkers[cls](self, node)

        if nid in self._seen_ids:
            self._seen_ids.remove(nid)
        return node

    def stop(self):
        raise StopIteration()

    #------------------------------------------------------
    # Generate walkers
    #------------------------------------------------------

    @classmethod
    def generate_walkers(cls):
        dispatch: Dict[Type[Node], Callable] = {}

        for node in WALK_FIELDS.keys():
            base_name = node.__name__.lower()
            enter_name = f"enter_{base_name}"
            exit_name = base_name if not keyword.iskeyword(base_name) else f"{base_name}_"

            # default no-op enter handler
            def _no_op(self, node):
                return None
            setattr(cls, enter_name, _no_op)
            setattr(cls, exit_name, _no_op)

            # Generate the per-class walker via exec for speed
            ns = {"NO_WALK": NO_WALK}
            lines =      ["def walker(self, node):"]
            lines.append(f"    no_walk = self.{enter_name}(node)")
            lines.append(f"    if no_walk is not NO_WALK:")
            lines.extend(f"        self._walk(node.{fld})" for fld in WALK_FIELDS.get(node, ()))
            lines.append(f"        self.{exit_name}(node)")
            exec("\n".join(lines), ns)
            dispatch[node] = ns["walker"] # type: ignore

        cls._walkers = dispatch

Walker.generate_walkers()

#------------------------------------------------------
# Rewriter
#------------------------------------------------------

class Rewriter:
    _rewriters: Dict[Type[Node], Callable] = {}
    __slots__ = ("_seen_ids", "_memo")

    def __init__(self):
        # ids of nodes on the current recursion path (for cycle detection)
        self._seen_ids: set[int] = set()
        # id(node) -> rewritten node
        self._memo: Dict[int, Node] = {}

    def __call__(self, obj):
        self._seen_ids.clear()
        self._memo.clear()
        return self._rewrite(obj)

    def _rewrite(self, x):
        if isinstance(x, Node):
            return self._rewrite_node(x)
        if isinstance(x, _ATOMS):
            return x
        if isinstance(x, (tuple, list)):
            seq_out: list[Any] = []
            changed = False
            for v in x:
                nv = self._rewrite(v)
                if nv is not v:
                    changed = True
                seq_out.append(nv)
            if not changed:
                return x
            return tuple(seq_out) if isinstance(x, tuple) else seq_out
        if isinstance(x, dict):
            out: Dict[Any, Any] = {}
            changed = False
            for k, v in x.items():
                nv = self._rewrite(v)
                if nv is not v:
                    changed = True
                out[k] = nv
            if not changed:
                return x
            return out
        return x

    def _rewrite_node(self, node: Node):
        nid = id(node)

        # If we've already rewritten this exact node, reuse the result.
        rewritten = self._memo.get(nid)
        if rewritten is not None:
            return rewritten

        # Cycle guard by identity for pathological graphs.
        if nid in self._seen_ids:
            return node

        self._seen_ids.add(nid)
        cls = type(node)
        rewriter = self._rewriters[cls]
        new = rewriter(self, node)
        self._seen_ids.remove(nid)

        self._memo[nid] = new
        return new

    #------------------------------------------------------
    # Generate rewriters
    #------------------------------------------------------

    @classmethod
    def generate_rewriters(cls):
        dispatch: Dict[Type[Node], Callable] = {}

        for node in WALK_FIELDS.keys():
            base_name = node.__name__.lower()
            enter_name = f"enter_{base_name}"
            exit_name = base_name if not keyword.iskeyword(base_name) else f"{base_name}_"

            # default handlers: return None => "no replacement"
            def _no_op(self, nd):
                return None
            setattr(cls, enter_name, _no_op)
            setattr(cls, exit_name, _no_op)

            fields = WALK_FIELDS.get(node, ())

            ns: Dict[str, Any] = {}
            lines: list[str] = []

            # def rewriter(self, node):
            lines.append("def rewriter(self, node):")
            lines.append("    new = node")

            # pre-order hook: enter_<name>(node) -> optional replacement
            lines.append(f"    tmp = self.{enter_name}(node)")
            lines.append(f"    if tmp is not None and tmp is not node:")
            lines.append(f"        new = tmp")
            # rewrite children in WALK_FIELDS, track changes
            lines.extend(f"    v_{fld} = self._rewrite(new.{fld})" for fld in fields)

            if fields:
                cond = " or ".join(f"v_{fld} is not new.{fld}" for fld in fields)
                kwargs = ", ".join(f"{fld}=v_{fld}" for fld in fields)
                lines.append(f"    if {cond}:")
                lines.append(f"        new = new.mut({kwargs})")

            # post-order hook: <name>(node) -> optional replacement
            lines.append(f"    tmp = self.{exit_name}(new)")
            lines.append(f"    if tmp is not None and tmp is not new:")
            lines.append(f"        new = tmp")

            lines.append("    return new")

            exec("\n".join(lines), {}, ns)
            dispatch[node] = ns["rewriter"]

        cls._rewriters = dispatch

Rewriter.generate_rewriters()



import typing
import v0.relationalai

import v0.relationalai.experimental
import v0.relationalai.experimental.pathfinder
import v0.relationalai.experimental.pathfinder.bridge
from v0.relationalai.experimental.pathfinder.rpq import RPQ

__all__ = [
    'node_filter',
    'node',
    'edge',
    'path',
    'optional',
    'plus',
    'star',
    'union',
    'RPQ',
    'find_paths'
]

#-----------------------------------------------------------------------------
# Parsing RPQ elements specified with strings (node and edge labels).
#-----------------------------------------------------------------------------

Segment = typing.Union[str, v0.relationalai.dsl.Type, RPQ]

def _process_segment(s: Segment) -> RPQ:
    from v0.relationalai.experimental.pathfinder.filter import NodeLabel, parse_label
    from v0.relationalai.experimental.pathfinder.rpq import Node, Edge

    if isinstance(s, str):
        label = parse_label(s)
        if isinstance(label, NodeLabel):
            return Node(label)
        else:
            return Edge(label)
    elif isinstance(s, v0.relationalai.dsl.Type):
        return Node(NodeLabel(s._type.name))
    elif isinstance(s, RPQ):
        return s
    else:
        raise Exception(f"Incorrect type {type(s)} of segment '{s}'")


#-----------------------------------------------------------------------------
# User-facing RPQ constructors
#-----------------------------------------------------------------------------

# (filter)
def node_filter(filter: typing.Callable) -> RPQ:
    from v0.relationalai.experimental.pathfinder.filter import AnonymousNodeFilter
    from v0.relationalai.experimental.pathfinder.rpq import Node

    return Node(label = None, filter = AnonymousNodeFilter(filter))


# (node_lab {∧ filter})
def node(node_lab: typing.Union[str,v0.relationalai.dsl.Type],
         filter: typing.Optional[typing.Callable] = None) -> RPQ:
    from v0.relationalai.experimental.pathfinder.filter import (
        NodeLabel, AnonymousNodeFilter, parse_label
    )
    from v0.relationalai.experimental.pathfinder.rpq import Node

    filter = AnonymousNodeFilter(filter) if filter is not None else None

    if isinstance(node_lab, v0.relationalai.dsl.Type):
        label = NodeLabel(node_lab._type.name)
    else:
        label = parse_label(node_lab)
    assert isinstance(label, NodeLabel), f"Expected a node label, got {label}"

    return Node(label, filter)

# -[edge_lab {∧ filter}]- or <-[edge_lab {∧ filter}]-
def edge(edge_lab:str, filter: typing.Optional[typing.Callable] = None) -> RPQ:
    from v0.relationalai.experimental.pathfinder.filter import (
        EdgeLabel, AnonymousEdgeFilter, parse_label
    )
    from v0.relationalai.experimental.pathfinder.rpq import Edge

    label = parse_label(edge_lab)
    assert isinstance(label, EdgeLabel), f"Expected an edge label, either `-[...]->` or `<-[...]-`, got {label}"
    filter = AnonymousEdgeFilter(filter) if filter is not None else None
    return Edge(label, filter)


# (expr₁ ⋅ ... ⋅ exprₙ)
def path(segment: Segment, *segments: Segment) -> RPQ:
    from v0.relationalai.experimental.pathfinder.rpq import Concat

    pattern = _process_segment(segment)
    for seg in segments:
        pattern = Concat(pattern, _process_segment(seg))

    return pattern

# (expr₁ ⋅ ... ⋅ exprₙ)?
def optional(*segments: Segment) -> RPQ:
    from v0.relationalai.experimental.pathfinder.rpq import Optional

    return Optional(path(*segments))

# (expr₁ ⋅ ... ⋅ exprₙ)+
def plus(segment: Segment, *segments: Segment) -> RPQ:
    from v0.relationalai.experimental.pathfinder.rpq import Plus

    return Plus(path(segment, *segments))

# (expr₁ ⋅ ... ⋅ exprₙ)*
def star(segment: Segment, *segments: Segment) -> RPQ:
    from v0.relationalai.experimental.pathfinder.rpq import Star

    return Star(path(segment, *segments))

# (expr₁ | ... | exprₙ)
def union(segment: Segment, *segments: Segment) -> RPQ:
    from v0.relationalai.experimental.pathfinder.rpq import Union

    pattern = _process_segment(segment)
    for seg in segments:
        pattern = Union(pattern, _process_segment(seg))

    return pattern

#-----------------------------------------------------------------------------
# UI for path finding
#-----------------------------------------------------------------------------

# Create a type storing paths specified by the given RPQ pattern.
def find_paths(pattern: RPQ, **params):
    from v0.relationalai.experimental.pathfinder.options import normalized_options
    from v0.relationalai.experimental.pathfinder.bridge import invoke_pathfinder

    options = normalized_options(params)

    return invoke_pathfinder(pattern, options)

# Creates a connectivity relation from a given RPQ pattern.
def conn(pattern: RPQ, **params):
    from v0.relationalai.std import rel
    from v0.relationalai.dsl import create_vars
    from v0.relationalai.experimental.pathfinder.options import normalized_options
    from v0.relationalai.experimental.pathfinder.compiler import compile_conn
    from v0.relationalai.experimental.pathfinder.datalog import install_program
    from v0.relationalai.experimental.pathfinder.utils import get_model

    options = {**normalized_options(params), 'suppress_groundedness_test': True}
    program = compile_conn(pattern, options)
    pq_conn_rel = program.root_rel['conn_rel']

    install_program(get_model(options), program, options)

    x, y = create_vars(2)
    getattr(rel, pq_conn_rel)(x, y)

    return x, y



import warnings

from v0.relationalai.semantics.std.strings import (string, len, contains, substring, like, lower, upper, strip, startswith,
                                                endswith, levenshtein, concat)

__all__ = ["string", "len", "contains", "substring", "like", "lower", "upper", "strip", "startswith", "endswith",
           "levenshtein", "concat"]

warnings.warn(
    "relationalai.early_access.builder.std.strings is deprecated, "
    "Please migrate to relationalai.semantics.std.strings",
    DeprecationWarning,
    stacklevel=2,
)


from enum import Enum
from typing import Any
from typing import Optional


class ProductStage(Enum):
    """Product stages for features and APIs."""
    GA = "generally available"
    PREVIEW = "public preview"
    EARLY_ACCESS = "early access"


def include_in_docs(
    obj: Any,
    stage: Optional[ProductStage] = None
) -> Any:
    """Marks an object for inclusion in documentation.

    Parameters
    ----------
    obj : Any
        The object to mark for inclusion in documentation.
    stage : ProductStage or None
        The stage of the product for which the object is included in documentation.
        Defaults to None, which means it inherits the product stage of its parent.

    Examples
    --------
    Use as a decorator to mark a function for inclusion in documentation:

    >>> @include_in_docs
    >>> def double(x: int):
    >>>     return x * 2

    Use as a decorator to mark classes and methods for inclusion in documentation:

    >>> @include_in_docs
    >>> class Dog:
    >>>     def __init__(self, name: str):
    >>>         self.name = name
    >>>     @include_in_docs(ProductStage.PREVIEW)
    >>>     def bark(self):
    >>>         return "Woof!"

    Use as a function to mark other objects like constants or modules for inclusion in documentation:

    >>> from my_package import my_module
    >>> include_in_docs(my_module, ProductStage.EARLY_ACCESS)
    >>> MY_CONSTANT = include_in_docs(42)
    """
    setattr(obj, "__include_in_docs__", True)
    setattr(obj, "__rai_product_stage__", stage)
    return obj
